# https://code.activestate.com/recipes/577197-sortedcollection/
import collections
import pprint
from bisect import bisect_left, bisect_right

# from .logger import logger


class Multimap(object):
    '''
    Python equivalent of C++ stl::Multimap

     "Multimap is an associative container that contains a sorted list of key-value pairs,
     while permitting multiple entries with the same key. Sorting is done according to
     the comparison function Compare, applied to the keys. Search, insertion, and removal
     operations have logarithmic complexity." -- https://en.cppreference.com/w/cpp/container/multimap


    * Member classes
        value_compare   compares objects of type value_type  [not implemented]
    * Member functions
        (constructor)   constructs the map
        (destructor)    destructs the map
        operator=       assigns values to the container   [not implemented]
        get_allocator   returns the associated allocator  [not implemented]

    * Iterators
        begin,cbegin    returns an iterator to the beginning   [__iter__]
        end,cend        returns an iterator to the end
        rbegin,crbegin  returns a reverse iterator to the beginning
        rend,crend      returns a reverse iterator to the end
    * Capacity
        empty           checks whether the container is empty
        size            returns the number of elements  [__len__]
        max_size        returns the maximum possible number of elements
    * Modifiers
        clear           clears the contents
        insert          inserts elements or nodes (since C++17)  [__setitem__]
        emplace         (C++11) constructs element in-place  [not implemented]
        emplace_hint    (C++11) constructs elements in-place using a hint  [not implemented]
        try_emplace     (C++17) inserts in-place if the key does not exist, does nothing if the key exists [not implemented]
        erase           erases elements [__delitem__]
        swap            swaps the contents [not implemented]
        extract         (C++17) extracts nodes from the container [not implemented]
        merge           (C++17) splices nodes from another container [not implemented]
    * Lookup
        count           returns the number of elements matching specific key
        find            finds element with specific key
        contains        checks if the container contains element with specific key  [__contains__]
        equal_range     returns range of elements matching a specific key
        lower_bound     returns an iterator to the first element not less than the given key
        upper_bound     returns an iterator to the first element greater than the given key
    * Observers
        key_comp        returns the function that compares keys [not implemented]
        value_comp      returns the function that compares keys in objects of type value_type [not implemented]
    * Non-member functions
        operator==      lexicographically compares the values in the map [not implemented]
        operator!=
        operator<
        operator<=
        operator>
        operator>=

        std::swap(std::Multimap) specializes the std::swap algorithm (function template)  [not implemented]
        erase_if(std::map)  (C++20) Erases all elements satisfying specific criteria (function template)
    
    '''

    def __init__(self, data=None, key=None, ** kwargs):
        self._key = key or (lambda x: x[0])
        self._keys = []
        self._items = []
        self.insert(data)

    # Capacity
    def __repr__(self):
        return pprint.pformat(self._items)

    def __str__(self):
        return pprint.pformat(self._items)

    def empty(self) -> bool:
        return self.size() == 0

    def size(self) -> int:
        return len(self._items)

    def __len__(self):
        return self.size()

    def items(self):
        return self._items

    def values(self):
        return self._items

    def keys(self):
        return self._keys

    # Modifiers

    def clear(self):
        self.__init__([], self._key)

    def insert_many(self, kvs, before=True):
        if before:
            op = self.insert_before
        else:
            op = self.insert_after
        if isinstance(kvs, collections.abc.Mapping):
            self.insert_many(kvs.items(), before)
        else:
            for kv in kvs:
                op(kv)
        # else:
        #     raise TypeError(f"illegal argument type! {type(kvs)}")

    def insert_before(self, kv):
        'Insert a new item.  If equal keys are found, add to the left'
        k = self._key(kv)
        i = bisect_left(self._keys, k)
        self._keys.insert(i, k)
        self._items.insert(i, kv)

    def insert_after(self, kv):
        'Insert a new item.  If equal keys are found, add to the right'
        k = self._key(kv)
        i = bisect_right(self._keys, k)
        self._keys.insert(i, k)
        self._items.insert(i, kv)

    def insert(self, kv, before=False):
        if kv is None:
            return
        if before:
            op = self.insert_before
        else:
            op = self.insert_after
        op(kv)

    def remove(self, arg0, arg1=None):
        raise NotImplementedError("erase")

    # Lookup
    def count(self, k):
        i = bisect_left(self._keys, k)
        j = bisect_right(self._keys, k)
        return j - i

    def contains(self, k):
        return self.count(k) > 0

    def __contains__(self, k):
        return self.count(k) > 0

    def find(self, k):
        idx = bisect_left(self._keys, k)
        if idx < len(self._keys) and self._keys[idx] == k:
            return self._items[idx]
        else:
            return None

    def find_lower(self, key):
        return self[:self.upper_bound(key)]

    def find_upper(self, key):
        return self[self.lower_bound(key):]

    # class Iterator(collections.abc.Iterator,  collections.abc.ItemsView):
    #     def __init__(self, container, start, end=None, step=1):
    #         self._container = container
    #         self._max = len(self._container)
    #         self._min = 0
    #         self._start = start
    #         self._end = end
    #         self._reverse = reverse
    #         self._it = self._end if self._reverse else self._start

    #     def reverse(self):
    #         return Multimap.Iterator(self._container, self._start, self._end, not self._reverse)

    #     def __iter__(self):
    #         return self

    #     def __next__(self):
    #         if self._reverse and (self._it >= self._start):
    #             self._it = self._it - 1
    #             return self._container[self._it]
    #         elif (not self._reverse) and self._it < self._end:
    #             idx = self._it
    #             self._it = self._it + 1
    #             return self._container[idx]
    #         else:
    #             raise StopIteration()

    #     def __contains__(self, k):
    #         idx = bisect_left(self._container._keys, k,
    #                           self._start, self._end)
    #         return idx >= self._start and idx < self._end

    def lower_bound(self, k):
        '''bisec_left: The return value i is such that all e in a[:i] have e < x,
         and all e in a[i:] have e >= x. So if x already appears in the list,
          i points just before the leftmost x already there.'''
        return bisect_left(self._keys, k)

    def upper_bound(self, k):
        '''bisect_right: The return value i is such that all e in a[:i] have e <= x, and all e in a[i:]
         have e > x. So if x already appears in the list, i points just beyond the rightmost x already there
        '''
        return bisect_right(self._keys, k)

    def equal_range(self, k):
        '''returns all elements that key == k'''
        lo = bisect_left(self._keys, k)
        hi = bisect_right(self._keys, k)

        return self._items[lo:hi]

    def reverse_equal_range(self, k):
        '''returns all elements that key == k'''
        lo = bisect_left(self._keys, k)
        hi = bisect_right(self._keys, k)

        return self._items[hi:lo][::-1]

    def lower_range(self, k):
        '''returns all elements that key < k'''
        return self._items[:bisect_left(self._keys, k)]

    def upper_range(self, k):
        '''returns all elements that key > k'''
        return self._items[:bisect_right(self._keys, k)]

    def range(self, k_lo, k_hi):
        '''returns all elements that  k_lo<= key < k_hi'''

        lo = bisect_left(self._keys, k_lo)
        hi = bisect_left(self._keys, k_hi)

        return self._items[lo:hi]

    ###############################################

    def copy(self):
        return self.__class__(self, self._key)

    def __getitem__(self, i):
        return self._items[i]

    def __iter__(self):
        return iter(self._items)

    # def __repr__(self):
    #     return f"{self.__class__.__name__}({pprint.pformat(self._items)}, key={ getattr(self._key, '__name__', repr(self._key))})"

    # def __str__(self):
    #     return f"{self.__class__.__name__}({pprint.pformat(self._items)}, key={ getattr(self._key, '__name__', repr(self._key))})"



import collections
import json

import numpy as np
from spdm.utils.logger import logger

from spdm.core.Collection import FileCollection
from spdm.core.Document import Document


class NumpyEncoder(json.NumPyEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()

        logger.debug(type(obj))
        return super().default(obj)


class NumPyDocument(Document):
    def __init__(self, path, *args, **kwargs):
        super().__init__(path, *args, **kwargs)
        self._path = path

    def load(self, *args, path=None,  **kwargs):
        with open(path or self._path, mode="r") as fid:
            self.root._holder = np.load(fid)

    def save(self,   *args, path=None, **kwargs):
        with open(path or self._path, mode="w") as fid:
            np.save(fid, self.root._holder)


class NumPyCollection(FileCollection):
    def __init__(self, uri, *args, **kwargs):
        super().__init__(uri, *args,
                         file_extension=".json",
                         file_factory=lambda *a, **k: NumPyDocument(*a, **k),
                         ** kwargs)



import typing
import numpy as np
from .Mesh import Mesh
from .mesh_structured import StructuredMesh
from ..utils.typing import ArrayType


@Mesh.register("uniform")
class UniformMesh(StructuredMesh):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        p_min, p_max = self.geometry.bbox
        self._dx = (np.asarray(p_max, dtype=float)-np.asarray(p_min, dtype=float))/np.asarray(self.shape, dtype=float)
        self._orgin = np.asarray(p_min, dtype=float)

    @property
    def origin(self) -> typing.Tuple[float]: return self._dx

    @property
    def dx(self) -> typing.Tuple[float]: return self._dx

    def vertices(self, *args) -> ArrayType:
        if len(args) == 1:
            uvw = np.asarray(uvw, dtype=float)
        else:
            uvw = np.stack(list(args))

        return np.stack([(uvw[i]*self.dx[i]+self.origin[i]) for i in range(self.rank)])





from __future__ import annotations

import collections
import collections.abc
import pathlib
import typing

import numpy as np
from lxml.etree import Comment as _XMLComment
from lxml.etree import ParseError as _XMLParseError
from lxml.etree import XPath as _XPath
from lxml.etree import _Element as _XMLElement
from lxml.etree import Element as Element

from lxml.etree import fromstring, tostring
from lxml.etree import parse as parse_xml
from spdm.core.Entry import Entry
from spdm.core.File import File
from spdm.core.Path import Path, PathLike
from spdm.utils.logger import logger
from spdm.utils.misc import normalize_path, serialize
from spdm.utils.PathTraverser import PathTraverser
from spdm.utils.tags import _not_found_, _undefined_
from spdm.utils.tree_utils import format_string_recursive


def merge_xml(first, second):
    if first is None:
        raise ValueError(f"Try merge to None Tree!")
    elif second is None:
        return first
    elif first.tag != second.tag:
        raise ValueError(
            f"Try to merge tree to different tag! {first.tag}<={second.tag}")

    for child in second:
        if child.tag is _XMLComment:
            continue
        eid = child.attrib.get("id", None)
        if eid is not None:
            target = first.find(f"{child.tag}[@id='{eid}']")
        else:
            target = first.find(child.tag)
        if target is not None:
            merge_xml(target, child)
        else:
            first.append(child)


def load_xml(path: str | list | pathlib.Path, *args,  mode: File.Mode | str = "r", **kwargs):
    # TODO: add handler non-local request ,like http://a.b.c.d/babalal.xml

    if isinstance(path, list):
        root = None
        for fp in path:
            if root is None:
                root = load_xml(fp, mode=mode)
            else:
                merge_xml(root, load_xml(fp, mode=mode))
        return root

    if isinstance(path, str):
        path = pathlib.Path(path)

    root = None
    if path.exists() and path.is_file():
        try:
            if isinstance(path, pathlib.Path):
                root = parse_xml(path.as_posix()).getroot()
            elif isinstance(path, str):
                root = parse_xml(path).getroot()
            else:
                raise TypeError(f"Invalid path type: {type(path)}")
            # logger.debug(f"Loading XML file from {path}")
        except _XMLParseError as msg:
            raise RuntimeError(f"ParseError: {path}: {msg}")
    else:
        raise FileNotFoundError(path)

    if root is not None:
        for child in root.findall("{http://www.w3.org/2001/XInclude}include"):
            fp = path.parent/child.attrib["href"]
            root.insert(0, load_xml(fp))
            root.remove(child)

    return root


def tree_to_xml(root: str | Element, d, *args, **kwargs) -> _XMLElement:
    if isinstance(root, str):
        root = Element(root)

    for key, val in d.items():
        if isinstance(val, dict):
            child = tree_to_xml(key, val)
            root.append(child)
        elif isinstance(val, list):
            for i, v in enumerate(val):
                child = tree_to_xml(key, v)
                child.set("id", str(i))
                root.append(child)

        elif isinstance(val,   np.ndarray):
            child = Element(key)
            if np.issubdtype(val.dtype, np.floating):
                child.set("dtype", "float")
            elif np.issubdtype(val.dtype, np.integer):
                child.set("dtype", "int")
            else:
                logger.debug(val.dtype)
            child.set("shape", str(list(val.shape)).strip("[]"))
            text = str(val.tolist())
            text = text.strip("[]")

            child.text = text
            root.append(child)
        else:
            child = Element(key)
            child.text = str(val)
            root.append(child)

    return root


@Entry.register("xml")
class XMLEntry(Entry):
    def __init__(self, data: _XMLElement | str, *args, envs={}, **kwargs):
        super().__init__({}, *args, **kwargs)

        if not isinstance(data, str):
            pass
        elif not data.strip(" ").startswith("<"):
            data = load_xml(data)
        else:
            data = fromstring(data)

        self._data: _XMLElement = data

        self._envs = envs

    def __repr__(self) -> str:
        # return f"<{self.__class__.__name__} root={self._root} path={self._path} />"
        return f"<{self._data.tag}  path=\"{self._path}\" />"

    def __copy__(self) -> Entry:
        other: XMLEntry = super().__copy__()  # type:ignore
        other._envs = self._envs
        other._data = self._data
        return other

    def _xpath(self, path):
        envs = {}
        res = "."
        prev = None
        for p in path:
            if isinstance(p, int):
                res += f"[(position()= {p+1} and @id ) or (@id={p}) or @id='*']"
                envs[prev] = p
            # # elif isinstance(p, slice):
            # #     if p == slice(None):
            # #         res += "[@id]"
            # #     else:
            # #         raise NotImplementedError("XML DO NOT SUPPORT SLICE!")
            # elif isinstance(p, (tuple, set)):
            #     raise NotImplementedError(f"XML DO NOT SUPPORT TUPLE OR SET!{path}")
            elif isinstance(p, str) and len(p) > 0:
                if p[0] == '@':
                    res += f"[{p}]"
                else:
                    res += f"/{p}"
                    prev = p
            else:
                envs[prev] = p
                # # TODO: handle slice
                # raise TypeError(f"Illegal path type! {type(p)} {path}")

        return res, envs

    def xpath(self, path):
        p, e = self._xpath(path)
        return _XPath(p), e

    def _convert(self, element: _XMLElement | list, path=[], lazy=False, envs=None, only_one=False, default_value: typing.Any = _not_found_, **kwargs):
        if not isinstance(element, list):
            pass

        elif len(element) == 0:
            return default_value

        elif len(path) > 0 and isinstance(path[-1], slice):
            raise NotImplementedError(f"{path}")

        else:
            res = [self._convert(e, path=path, lazy=lazy, envs=envs, **kwargs) for e in element]

            if only_one:
                res = res[0]
            elif len(res) == 1:
                res = res[0]

            return res

        res = None
        text = element.text.strip() if element.text is not None else None
        if text is not None and len(text) > 0:
            if "dtype" in element.attrib or (len(element) == 0 and len(element.attrib) == 0):
                dtype = element.attrib.get("dtype", None)
                if dtype == "string" or dtype is None:
                    res = [text]
                elif dtype == "int":
                    res = [int(v.strip()) for v in text.strip(',').split(',')]
                elif dtype == "float":
                    res = [float(v.strip()) for v in text.strip(',').split(',')]
                else:
                    raise NotImplementedError(f"Not supported dtype {dtype}!")

                dims = [int(v) for v in element.attrib.get("dims", "").split(',') if v != '']
                if len(dims) == 0 and len(res) == 1:
                    res = res[0]
                elif len(dims) > 0 and len(res) != 0:
                    res = np.array(res).reshape(dims)
                else:
                    res = np.array(res)
            elif len(element.attrib) == 0:
                res = text
            else:
                res = {}
                for k, v in element.attrib.items():
                    res[f"@{k}"] = v
                res["_text"] = text

        elif not lazy:
            res = {}
            for child in element:
                if child.tag is _XMLComment:
                    continue
                obj = self._convert(
                    child, path=path+[child.tag], envs=envs, lazy=lazy, **kwargs)
                tmp = res.setdefault(child.tag, obj)
                if tmp is obj:
                    continue
                elif isinstance(tmp, list):
                    tmp.append(obj)
                else:
                    res[child.tag] = [tmp, obj]

            # res = {child.tag: self._convert(child, path=path+[child.tag], envs=envs, lazy=lazy, **kwargs)
            #        for child in element if child.tag is not _XMLComment}
            for k, v in element.attrib.items():
                res[f"@{k}"] = v

            text = element.text.strip() if element.text is not None else None
            if text is not None and len(text) != 0:
                query = {}
                prev = None
                for p in self._prefix+path:
                    if type(p) is int:
                        query[f"{prev}"] = p
                    prev = p

                # if not self._envs.fragment:
                #     fstr = query
                # else:
                #     fstr = collections.ChainMap(query, self.envs.fragment.__data__, self.envs.query.__data__ or {})
                # format_string_recursive(text, fstr)  # text.format_map(fstr)
                res["@text"] = text
        else:
            res = XMLEntry(element, prefix=[], envs=envs)

        if envs is not None and isinstance(res, (str, collections.abc.Mapping)):
            res = format_string_recursive(res, collections.ChainMap(envs, self._envs))

        if not isinstance(path[-1], int) and isinstance(res, dict) and res.get("@id", None) == "*":
            logger.debug((res.get("@id", None), path))
            if not isinstance(path[-1], int):
                res = [res]

        return res

    #############################
    # API

    def insert(self,  *args, **kwargs) -> XMLEntry: raise NotImplementedError(f"")

    def update(self,  *args, **kwargs) -> XMLEntry: raise NotImplementedError(f"")

    def remove(self,  *args, **kwargs) -> int: raise NotImplementedError(f"")

    def find(self, op=None, *args, **kwargs) -> typing.Any:

        path = self._path[:]
        xp, envs = self.xpath(path)

        obj: typing.List[_XMLElement] = xp.evaluate(self._data)

        if op is Path.tags.exists:
            return len(obj) > 0
        elif op is Path.tags.count:
            return len(obj)
        elif op is Path.tags.is_leaf:
            if len(obj) == 0:
                return _not_found_
            elif len(obj) > 1:
                return False
            else:
                return len(obj[0]) == 0

        elif op is Path.tags.is_list:
            return len(obj) > 1 or (len(obj) == 1 and obj[0].attrib.get("id", None) != None)

        elif op is Path.tags.is_dict:
            return len(obj) == 1

        elif op is None or op is Path.tags.find:
            return self._convert(obj, path=path,   envs=envs, **kwargs)
        else:
            target = self._convert(obj, path=path,   envs=envs, **kwargs)

            return Path._apply_op(target, op, [], *args)

    def search_next(self,  start=None,  **kwargs) -> typing.Tuple[typing.Any, int | None]:

        if len(self._path) > 0 and isinstance(self._path[-1], slice):
            start = self._path[-1].start or start or 0
            stop = self._path[-1].stop
            step = self._path[-1].step or 1
            path = self._path[:-1] + [start]

        elif isinstance(self._path[-1], dict):
            raise NotImplementedError(f"Can not search next element from dict!")

        else:
            if start is None or start is _not_found_:
                start = 0
            path = self._path[:] + [start]
            stop = None
            step = 1

        if stop is not None and start >= stop:
            raise StopIteration(f"{start}>{stop}")

        xp, envs = self.xpath(path)
        data = xp.evaluate(self._data)
        if len(data) == 0:
            raise StopIteration(f"Can not search next element from {path}")
        elif len(data) == 1:
            res = self._convert(data[0], lazy=True, path=path, envs=envs, **kwargs)
            return res, start+step
        else:
            raise RuntimeError(f"Invalid path {path}")

    def for_each(self, *args, **kwargs) -> typing.Generator[typing.Tuple[int, typing.Any], None, None]:
        """ Return a generator of the results. """
        if len(self._path) > 0 and isinstance(self._path[-1], slice):
            start = self._path[-1].start or 0
            stop = self._path[-1].stop
            step = self._path[-1].step or 1
            prefix = self._path[:-1]

        elif isinstance(self._path[-1], dict):
            raise NotImplementedError(f"Can not search next element from dict!")

        else:

            start = 0
            stop = None
            step = 1
            prefix = self._path[:]

        next_id = start
        while True:
            if stop is not None and next_id >= stop:
                raise StopIteration(f"{next_id}>{stop}")
            path = prefix+[next_id]
            xp, envs = self.xpath(prefix+[next_id])
            data = xp.evaluate(self._data)
            if len(data) == 0:
                # raise StopIteration(f"Can not search next element from {path}")
                break
            elif len(data) == 1:
                res = self._convert(data[0], lazy=True, path=path, envs=envs, **kwargs)
                yield next_id, res
                next_id += step
            else:
                raise RuntimeError(f"Invalid path {path}")

    #############################

    def _get_value(self,  path: PathLike = None, *args,  only_one=False, default_value=_not_found_, **kwargs):

        if not only_one:
            return PathTraverser(path).apply(lambda p: self._get_value(p, only_one=True, **kwargs))
        else:
            path = self._prefix+normalize_path(path)
            xp, envs = self.xpath(path)
            obj = xp.evaluate(self._data)
            if isinstance(obj, collections.abc.Sequence) and len(obj) == 1:
                obj = obj[0]
            return self._convert(obj, lazy=False, path=path, envs=envs, **kwargs)

    def search(self,  *args, envs={}, **kwargs):
        # path, s_envs = self._xpath(self._path[:])
        # TODO: PathTraverser(path):
        for s_path in self._path.traversal():
            s_path, s_envs = self._xpath(s_path)
            res = _XPath(s_path).evaluate(self._data)

            if len(res) == 0:
                break
            for child in res:
                if child.tag is not _XMLComment:
                    yield self._convert(child, path=s_path,
                                        envs=collections.ChainMap(s_envs, envs))

    def items(self, *args, envs={}, **kwargs):
        path = self._path
        for spath in PathTraverser(path):
            xp, s_envs = self.xpath(spath)
            for child in xp.evaluate(self._data):
                if child.tag is _XMLComment:
                    continue
                res = self._convert(child, path=spath,
                                    envs=collections.ChainMap(s_envs, envs))
                yield child.tag, res

    def values(self, *args, envs={}, **kwargs):
        path = self._path
        for spath in PathTraverser(path):
            xp, s_envs = self.xpath(spath)
            for child in xp.evaluate(self._data):
                if child.tag is _XMLComment:
                    continue
                res = self._convert(child, path=spath,
                                    envs=collections.ChainMap(s_envs, envs))
                yield res

    @property
    def attribute(self): return self._data.attrib

    def __serialize__(self): return serialize(self.find(default_value=_not_found_))


@File.register(["xml"])
class FILEPLUGINxml(File):
    def __init__(self, *args, root=None, **kwargs):
        super().__init__(*args, ** kwargs)
        if self.mode == File.Mode.read:
            self._root = load_xml(self.url.path, mode=self.mode)
        else:
            self._root = Element(root or "root")

    def read(self, lazy=True) -> Entry:
        return XMLEntry(self._root, writable=False)

    def write(self, data, *args, **kwargs) -> None:
        tree_to_xml(self._root, data, *args, **kwargs)
        with open(self.url.path, "w") as f:
            f.write(tostring(self._root, pretty_print=True).decode())


__SP_EXPORT__ = FILEPLUGINxml



import unittest

import numpy as np
from scipy import constants
from spdm.core.Expression import Expression, Variable
from spdm.core.Function import Function
from spdm.numlib.calculus import antiderivative, derivative, partial_derivative
from spdm.utils.logger import logger

TWOPI = constants.pi*2.0


class TestCalculus(unittest.TestCase):

    def test_dydx(self):

        _x = Variable(0, "x")

        x = np.linspace(0, TWOPI, 512)

        Y = Function(np.sin(_x), x, periods=[TWOPI])

        self.assertTrue(np.allclose(np.sin(x), Y(x), rtol=1.0e-4))

        dY = Y.derivative()(x)

        self.assertTrue(np.allclose(np.cos(x), dY, rtol=1.0e-4))

        # logger.debug((-(TWOPI**2)*np.sin(x))[:10])
        # logger.debug(Y.d(2)(x)[:10])
        # self.assertTrue(np.allclose(-(TWOPI**2)*np.sin(x), Y.d(2)(x), rtol=0.10))

    def test_integral(self):

        x = np.linspace(0, TWOPI, 512)

        _x = Variable(0, "x")

        Y = Function(np.cos(_x), x, periods=[TWOPI])

        Y1 = Y.antiderivative()

        self.assertTrue(np.allclose(np.sin(x), Y1(x), rtol=1.0e-4))

    def test_spl2d(self):

        x = np.linspace(0, TWOPI, 128)
        y = np.linspace(0, 2*TWOPI, 128)
        g_x, g_y = np.meshgrid(x, y)

        z = np.sin(g_x)*np.cos(g_y)

        _x = Variable(0, "x")

        _y = Variable(1, "y")

        fun = Function(np.sin(_x)*np.cos(_y), x, y, periods=[TWOPI,  2*TWOPI])

        z2 = fun(g_x, g_y)

        self.assertTrue(np.allclose(z, z2, rtol=1.0e-4))

    def test_pd2(self):

        x = np.linspace(0, TWOPI, 128)
        y = np.linspace(0, 2*TWOPI, 128)

        g_x, g_y = np.meshgrid(x, y)

        _x = Variable(0, "x")
        _y = Variable(1, "y")

        Z = Function(np.sin(_x)*np.cos(_y),  x, y, periods=[TWOPI, 2*TWOPI])

        self.assertTrue(np.allclose(np.sin(g_x)*np.cos(g_y),  Z(g_x, g_y), rtol=1.0e-4))

        dZdx = Z.partial_derivative(1, 0)
        self.assertTrue(np.allclose(np.cos(g_x)*np.cos(g_y), dZdx(g_x, g_y), rtol=1.0e-4))

        # ignore boundary points
        self.assertTrue(np.allclose((- np.sin(g_x)*np.sin(g_y))
                        [2:-2, 2:-2],  Z.pd(0, 1)(g_x, g_y)[2:-2, 2:-2], rtol=1.0e-4))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing_extensions import Self
import numpy as np
import collections
import collections.abc
import dataclasses
import inspect
import os
import pathlib
import typing
from copy import copy, deepcopy

from ..utils.logger import deprecated, logger
from ..utils.plugin import Pluggable
from ..utils.tags import _not_found_, _undefined_
from ..utils.typing import array_type, as_array, as_value, is_scalar
from ..utils.uri_utils import URITuple, uri_split, uri_split_as_dict
from .Path import Path, PathLike, as_path, update_tree, update_tree

PROTOCOL_LIST = ["local", "file", "http", "https", "ssh", "mdsplus"]


class Entry(Pluggable):
    _plugin_prefix = "spdm.plugins.data.plugin_"
    _plugin_registry = {}

    def __init__(
        self,
        data: typing.Any = _not_found_,
        path: Path | PathLike = [],
        *args,
        scheme=None,
        **kwargs,
    ):
        if self.__class__ is not Entry:
            pass
        elif scheme is not None or isinstance(data, (str, URITuple, pathlib.Path)):
            if scheme is None:
                scheme = uri_split(data).protocol

            if isinstance(scheme, str) and scheme != "":
                super().__dispatch_init__([scheme, EntryProxy], self, data, *args, **kwargs)
                return

        self._data = data
        self._path = as_path(path)

    def __copy__(self) -> Self:
        other = object.__new__(self.__class__)
        other._data = self._data
        other._path = deepcopy(self._path)
        return other

    def reset(self, value=None, path=None) -> Entry:
        self._data = value
        self._path = as_path(path)
        return self

    def __str__(self) -> str:
        return f'<{self.__class__.__name__} path="{self._path}" />'

    def __repr__(self) -> str:
        return str(self._path)

    @property
    def __entry__(self) -> Entry:
        return self

    @property
    def is_writable(self) -> bool:
        return False

    @property
    def path(self) -> Path:
        return self._path

    @property
    def is_leaf(self) -> bool:
        return self.find(Path.tags.is_leaf)

    @property
    def is_list(self) -> bool:
        return self.find(Path.tags.is_list)

    @property
    def is_dict(self) -> bool:
        return self.find(Path.tags.is_dict)

    @property
    def is_root(self) -> bool:
        return len(self._path) == 0

    @property
    def is_generator(self) -> bool:
        return self._path.is_generator

    ###########################################################

    @property
    def _value_(self) -> typing.Any:
        return self._data if len(self._path) == 0 else self.get(default_value=_not_found_)

    def __getitem__(self, *args) -> Entry:
        return self.child(*args)

    def __setitem__(self, path, value):
        return self.child(path).update(value)

    def __delitem__(self, *args):
        return self.child(*args).remove()

    def get(self, query=None, default_value=_not_found_, **kwargs) -> typing.Any:
        if query is None:
            entry = self
            args = ()
        elif isinstance(query, (slice, set, dict)):
            entry = self
            args = (query,)
        else:
            entry = self.child(query)
            args = ()

        return entry.find(Path.tags.find, *args, default_value=default_value, **kwargs)

    def put(self, pth, value, *args, **kwrags) -> Entry:
        entry = self.child(pth)
        entry.update(value, *args, **kwrags)
        return entry

    def dump(self, *args, **kwargs) -> typing.Any:
        return self.find(Path.tags.dump, *args, **kwargs)

    def equal(self, other) -> bool:
        if isinstance(other, Entry):
            return self.find(Path.tags.equal, other.__value__)
        else:
            return self.find(Path.tags.equal, other)

    @property
    def count(self) -> int:
        return self.find(Path.tags.count)

    @property
    def exists(self) -> bool:
        return self.find(Path.tags.exists)

    def check_type(self, tp: typing.Type) -> bool:
        return self.find(Path.tags.check_type, tp)

    ###########################################################

    @property
    def root(self) -> Entry:
        other = copy(self)
        other._path = []
        return other

    @property
    def parent(self) -> Entry:
        other = copy(self)
        other._path = self._path.parent
        return other

    def child(self, path=None, *args, **kwargs) -> Entry:
        path = as_path(path)
        if len(path) == 0:
            return self

        if self._data is not None or len(self._path) == 0:
            pass
        elif isinstance(self._path[0], str):
            self._data = {}
        else:
            self._data = []

        other = self.__copy__()
        other._path.append(path)
        return other

    def next(self, inc: int = 1) -> Entry:
        if not isinstance(self._path[-1], int) and not np.issubdtype(type(self._path[-1]), np.integer):
            raise RuntimeError(f"Path must be end with int! {self._path[-1]} {type(self._path[-1])}")

        next_ = self.__copy__()

        next_._path[-1] += inc

        return next_

    ###########################################################
    # API: CRUD  operation

    def insert(self, value, *args, **kwargs) -> Entry:
        self._data, *others = self._path.insert(self._data, value, *args, **kwargs)
        return self.__class__(self._data, *others)

    def update(self, value, *args, **kwargs) -> Entry:
        self._data = self._path.update(self._data, value, *args, **kwargs)
        return self

    def remove(self, *args, **kwargs) -> int:
        self._data, num = self._path.remove(self._data, *args, **kwargs)
        return num

    def find(self, *args, **kwargs) -> typing.Any:
        """
        Query the Entry.
        Same function as `find`, but put result into a contianer.
        Could be overridden by subclasses.
        """
        return self._path.find(self._data, *args, **kwargs)

    def keys(self) -> typing.Generator[str, None, None]:
        yield from self._path.keys(self._data)

    def for_each(self, *args, **kwargs) -> typing.Generator[typing.Any, None, None]:
        """Return a generator of the results."""
        yield from self._path.for_each(self._data, *args, **kwargs)

    def search(self, *args, **kwargs) -> Entry:
        raise NotImplementedError()

    ###########################################################


class ChainEntry(Entry):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._entrys: typing.List[Entry] = [
            (_open_entry(v, **kwargs) if not isinstance(v, Entry) else v)
            for v in args
            if v is not _not_found_ and v is not _undefined_ and v is not None
        ]

    def __copy__(self) -> ChainEntry:
        other = super().__copy__()
        other._entrys = self._entrys
        return other

    def __str__(self) -> str:
        return ",".join([str(e) for e in self._entrys if e._data is None])

    @property
    def is_writable(self) -> bool:
        return self._entrys[0].is_writable

    def find(self, *args, default_value=_not_found_, **kwargs):
        if len(args) > 0 and args[0] is Path.tags.count:
            res = super().find(*args, default_value=_not_found_, **kwargs)
            if res is _not_found_ or res == 0:
                for e in self._entrys:
                    res = e.child(self._path).find(*args, default_value=_not_found_, **kwargs)
                    if res is not _not_found_ and res != 0:
                        break
        else:
            res = super().find(*args, default_value=_not_found_, **kwargs)

            if res is _not_found_:
                for e in self._entrys:
                    e_child = e.child(self._path)
                    res = e_child.find(*args, default_value=_not_found_, **kwargs)
                    if res is not _not_found_:
                        break

            if res is _not_found_:
                res = default_value

        return res

    def for_each(self, *args, **kwargs) -> typing.Generator[typing.Tuple[int, typing.Any], None, None]:
        """逐个遍历子节点，不判断重复 id

        Returns:
            typing.Generator[typing.Tuple[int, typing.Any], None, None]: _description_

        Yields:
            Iterator[typing.Generator[typing.Tuple[int, typing.Any], None, None]]: _description_
        """

        for root_entry in self._entrys:
            yield from root_entry.child(self._path).for_each(*args, **kwargs)

            # 根据第一个有效 entry 中的序号，在其他 entry 中的检索子节点

            # if not _entry.exists:
            #     continue
            # for k, e in _entry.for_each(*args, **kwargs):
            # 根据子节点的序号，在其他 entry 中的检索子节点
            # entry_list = [e]
            # for o in self._entrys[idx + 1 :]:
            #     t = o.child(k)
            #     if t.exists:
            #         entry_list.append(t)
            # yield k, ChainEntry(*entry_list)

    def search(self, *args, **kwargs):
        return ChainEntry(*[e.search(*args, **kwargs) for e in self._entrys])

    @property
    def exists(self) -> bool:
        res = [super().find(Path.tags.exists)]
        res.extend([e.child(self._path).find(Path.tags.exists) for e in self._entrys])
        return any(res)


def _open_entry(entry: str | URITuple | pathlib.Path | Entry, mapping_files=None, local_schema=None, **kwargs) -> Entry:
    """
    Open an Entry from a URL.

    Using urllib.urlparse to parse the URL.  rfc3986

    URL format: <protocol>://<authority>/<path>?<query>#<fragment>

    RF3986 = r"^((?P<protocol>[^:/?#]+):)?(//(?P<authority>[^/?#]*))?(?P<path>[^?#]*)(\\?(?P<query>[^#]*))?(#(?P<fragment>.*))?")

    Example:
        ../path/to/file.json                    => File
        file:///path/to/file                    => File
        ssh://a.b.c.net/path/to/file            => ???
        https://a.b.c.net/path/to/file          => ???

        imas+ssh://a.b.c.net/path/to/file

        east+mdsplus://<mds_prefix>
        east+mdsplus+ssh://<mds_prefix>

    """
    if isinstance(entry, (dict, list)):
        # 如果是一个字典或者列表，直接转换成 Entry
        return Entry(entry)

    elif isinstance(entry, Entry):
        # 如果是一个Entry，直接返回
        if len(kwargs) > 0:  # pragma: no cover
            logger.warning(f"ignore {kwargs}")
        return entry

    elif isinstance(entry, str):
        # 如果是一个字符串，需要进行解析
        uri = uri_split(entry)

    elif isinstance(entry, pathlib.Path):
        # 如果是一个pathlib.Path，需要转换
        uri = URITuple(protocol="file", fragment=kwargs.pop("fragment", ""), query=kwargs, path=entry)

    elif isinstance(entry, URITuple):
        # 如果是一个URITuple，不需要转换
        uri = entry

    else:
        # 其他类型，报错
        raise RuntimeError(f"Unknown entry {entry} {type(entry)}")

    # if not isinstance(uri.path, str):
    #     raise RuntimeError(f"{entry} {uri}")

    fragment = uri.fragment

    query = update_tree(uri.query, kwargs)

    global_schema = query.pop("global_schema", None)

    schemas = [s for s in uri.protocol.split("+") if s != ""]

    local_schema = local_schema or query.pop("local_schema", None) or query.pop("device", None)
    if len(schemas) > 0 and schemas[0] not in PROTOCOL_LIST:
        local_schema = schemas[0]
        schemas = schemas[1:]

    new_url = URITuple(
        protocol="+".join(schemas),
        authority=uri.authority,
        path=uri.path,
        query=query,
        fragment="",
    )

    if new_url.protocol.startswith(("local+", "file+")):  # or (new_url.protocol == "" and new_url.path != ""):
        # 单一文件不进行 schema 检查，直接读取。因为schema转换在文件plugin中进行。
        from .File import File

        entry = File(new_url, **query).read()
    elif new_url.protocol.startswith(("http", "https", "ssh")):
        # http/https/ssh 协议，不进行schema检查，直接读取
        raise NotImplementedError(f"{new_url}")
    elif local_schema is not None and global_schema != local_schema:
        # 本地schema和全局schema不一致，需要进行schema转换
        entry = EntryProxy(new_url, local_schema=local_schema, global_schema=global_schema, **query)
    else:
        entry = Entry(new_url, **query)

    if fragment:
        entry = entry.child(fragment.replace(".", "/"))

    return entry


def open_entry(entry, local_schema=None, **kwargs) -> Entry:
    if (entry is None or isinstance(entry, Entry)) and local_schema is None:
        if len(kwargs) > 0:
            logger.warning(f"ignore {kwargs}")
        return entry
    elif isinstance(entry, list) and len(entry) == 1 and isinstance(entry[0], Entry):
        return entry[0]

    if entry is None or entry is _not_found_ or (isinstance(entry, list) and len(entry) == 0):
        if local_schema is None:
            return None
        else:
            entry = [f"{local_schema}://"]
    if not isinstance(entry, list):
        entry = [entry]

    entry = [a for a in entry if a is not None and a is not _not_found_]

    if isinstance(local_schema, str) and not any(
        [e.startswith(f"{local_schema}+") or e.startswith(f"mdsplus://") for e in entry if isinstance(e, str)]
    ):
        # just a walk around for mdsplus://
        entry = [f"{local_schema}://"] + entry

    if len(entry) == 0:
        return None

    elif len(entry) > 1:
        return ChainEntry(*entry, local_schema=local_schema, **kwargs)

    else:
        return _open_entry(entry[0], local_schema=local_schema, **kwargs)

    # url = uri_split(url_s)

    # # scheme = url.protocol.split("+")

    # # local_schema = None

    # # global_schema = schema

    # # match len(scheme):
    # #     case 0:
    # #         url.protocol = 'local'
    # #     case 1:
    # #         if scheme[0] not in _predefined_protocols:
    # #             local_schema = scheme[0]
    # #             url.protocol = ""
    # #     case _:
    # #         if scheme[0] in _predefined_protocols:
    # #             url.protocol = scheme[0]
    # #             if scheme[1] != kwargs.setdefault("format", scheme[1]):
    # #                 raise ValueError(f"Format mismatch! {scheme[1]} != {kwargs['format']}")
    # #         else:
    # #             local_schema = scheme[0]
    # #             url.protocol = "+".join(scheme[1:])

    # # if local_schema is not None:
    # #     kwargs.update(url.query)
    # #     if url.authority != '' or url.path != '':
    # #         kwargs['netloc'] = url.authority
    # #         kwargs['path'] = url.path
    # #     return EntryProxy(local_schema=local_schema, global_schema=global_schema, ** kwargs)

    # if url.protocol in ["file", "local", "", None]:
    #     from .File import File
    #     return File(url, *args, **kwargs).read()

    # elif url.protocol in ["https", "http"]:
    #     return Entry(url, *args, **kwargs)

    # else:
    #     raise RuntimeError(f"Unknown url {url} {Entry._plugin_registry}")


def as_entry(obj, *args, **kwargs) -> Entry:
    if isinstance(obj, Entry):
        entry = obj
    elif isinstance(obj, (str, URITuple, pathlib.Path)):
        entry = open_entry(obj, *args, **kwargs)
    elif hasattr(obj.__class__, "__entry__"):
        entry = obj.__entry__
    elif obj is None or obj is _not_found_:
        entry = Entry()
    else:
        entry = Entry(obj, *args, **kwargs)

    return entry


def as_dataclass(dclass, obj, default_value=None):
    if dclass is dataclasses._MISSING_TYPE:
        return obj

    if hasattr(obj, "entry"):
        obj = obj.entry
    if obj is None:
        obj = default_value

    if obj is None or not dataclasses.is_dataclass(dclass) or isinstance(obj, dclass):
        pass
    # elif getattr(obj, 'empty', False):
    #   obj = None
    elif dclass is array_type:
        obj = as_array(obj)
    elif hasattr(obj.__class__, "get"):
        obj = dclass(
            **{
                f.name: as_dataclass(
                    f.type,
                    obj.get(
                        f.name,
                        f.default if f.default is not dataclasses.MISSING else None,
                    ),
                )
                for f in dataclasses.fields(dclass)
            }
        )
    elif isinstance(obj, collections.abc.Sequence):
        obj = dclass(*obj)
    else:
        try:
            obj = dclass(obj)
        except Exception as error:
            logger.debug((type(obj), dclass))
            raise error
    return obj


def deep_reduce(first=None, *others, level=-1):
    if level == 0 or len(others) == 0:
        return first if first is not _not_found_ else None
    elif first is None or first is _not_found_:
        return deep_reduce(others, level=level)
    elif isinstance(first, str) or is_scalar(first):
        return first
    elif isinstance(first, array_type):
        return sum([first, *(v for v in others if (v is not None and v is not _not_found_))])
    elif len(others) > 1:
        return deep_reduce(first, deep_reduce(others, level=level), level=level)
    elif others[0] is None or first is _not_found_:
        return first
    elif isinstance(first, collections.abc.Sequence):
        if isinstance(others[0], collections.abc.Sequence) and not isinstance(others, str):
            return [*first, *others[0]]
        else:
            return [*first, others[0]]
    elif isinstance(first, collections.abc.Mapping) and isinstance(others[0], collections.abc.Mapping):
        second = others[0]
        res = {}
        for k, v in first.items():
            res[k] = deep_reduce(v, second.get(k, None), level=level - 1)
        for k, v in second.items():
            if k not in res:
                res[k] = v
        return res
    elif others[0] is None or others[0] is _not_found_:
        return first
    else:
        raise TypeError(f"Can not merge dict with {others}!")


def convert_fromentry(cls, obj, *args, **kwargs):
    origin_type = getattr(cls, "__origin__", cls)
    if dataclasses.is_dataclass(origin_type):
        obj = as_dataclass(origin_type, obj)
    elif inspect.isclass(origin_type):
        obj = cls(obj, *args, **kwargs)
    elif callable(cls) is not None:
        obj = cls(obj, *args, **kwargs)

    return obj


SPDB_XML_NAMESPACE = "{http://fusionyun.org/schema/}"

SPDB_TAG = "spdb"
########################################################
#  mapping files 目录结构约定为 :
#         ```{text}
#         - <local schema>/<global schema>
#              config.xml
#             - static            # 存储静态数据，例如装置描述文件
#                 - config.xml
#                 - <...>

#             - protocol0         # 存储 protocol0 所对应mapping，例如 mdsplus
#                 - config.xml
#                 - <...>

#             - protocol1         # 存储 protocol1 所对应mapping，例如 hdf5
#                 - config.xml
#                 - <...>
#         ```
#         Example:   east+mdsplus://.... 对应的目录结构为
#         ```{text}
#         - east/imas/3
#             - static
#                 - config.xml
#                 - wall.xml
#                 - pf_active.xml  (包含 pf 线圈几何信息)
#                 - ...
#             - mdsplus
#                 - config.xml (包含<spdb > 描述子数据库entry )
#                 - pf_active.xml
#         ```


class EntryProxy(Entry):
    _maps = {}
    _mapping_path = []
    _default_local_schema: str = "EAST"
    _default_global_schema: str = "imas/3"

    @classmethod
    def load(
        cls,
        url: str | None = None,
        local_schema: str = None,
        global_schema: str = None,
        mapping_files=None,
        **kwargs,
    ):
        """检索并导入 mapping files"""
        from .File import File

        if len(EntryProxy._mapping_path) == 0:
            EntryProxy._mapping_path.extend(
                [pathlib.Path(p) for p in os.environ.get("SP_DATA_MAPPING_PATH", "").split(":") if p != ""]
            )

        mapper_list = EntryProxy._maps

        _url = uri_split(url)

        kwargs = update_tree(url.query, kwargs)

        enabled_entry = kwargs.pop("enable", "").split(",")

        if local_schema is None:
            local_schema = EntryProxy._default_local_schema

        if global_schema is None:
            global_schema = EntryProxy._default_global_schema

        map_tag = [local_schema.lower(), global_schema.lower()]

        if _url.protocol != "":
            map_tag.append(_url.protocol)

        map_tag_str = "/".join(map_tag)

        mapper = mapper_list.get(map_tag_str, _not_found_)

        if mapper is _not_found_:
            prefix = "/".join(map_tag[:2])

            config_files = [
                f"{prefix}/config.xml",
                f"{prefix}/static/config.xml",
                f"{prefix}/{local_schema.lower()}.xml",
            ]

            if len(map_tag) > 2:
                config_files.append(f"{'/'.join(map_tag[:3])}/config.xml")

            if mapping_files is None:
                mapping_files: typing.List[pathlib.Path] = []

            for m_dir in EntryProxy._mapping_path:
                if not m_dir:
                    continue
                elif isinstance(m_dir, str):
                    m_dir = pathlib.Path(m_dir)

                for file_name in config_files:
                    p = m_dir / file_name
                    if p.exists():
                        mapping_files.append(p)

            if len(mapping_files) == 0:
                raise FileNotFoundError(
                    f"Can not find mapping files for {map_tag} MAPPING_PATH={EntryProxy._mapping_path} !"
                )

            mapper = File(mapping_files, mode="r", format="XML").read()

            mapper_list[map_tag_str] = mapper

        entry_list = {}

        spdb = mapper.child("spdb").find()

        if not isinstance(spdb, dict):
            entry_list["*"] = _url
        else:
            attr = {k[1:]: v for k, v in spdb.items() if k.startswith("@")}

            attr["prefix"] = f"{_url.protocol}://{_url.authority}{_url.path}"

            attr.update(kwargs)

            for entry in spdb.get("entry", []):
                id = entry.get("@id", None)

                enable = entry.get("@enable", "true") == "true"

                if id is None:
                    continue
                elif not enable and id not in enabled_entry:
                    continue

                entry_list[id] = entry.get("_text", "").format(**attr)

        return mapper, entry_list

    def __init__(self, *args, **kwargs):
        super().__init__()
        if len(args) == 2 and isinstance(args[0], Entry) and isinstance(args[1], dict):
            self._mapper = args[0]
            self._entry_list = args[1]
        else:
            self._mapper, self._entry_list = self.__class__.load(*args, **kwargs)

    def __copy__(self) -> Self:
        other = super().__copy__()
        other._mapper = self._mapper
        other._entry_list = self._entry_list
        return other

    def __str__(self) -> str:
        return ",".join([str(e) for e in self._entry_list.values() if isinstance(e, str)])

    def child(self, *args, **kwargs) -> Self:
        res: EntryProxy = super().child(*args, **kwargs)
        res._entry_list = self._entry_list
        return res

    def insert(self, value, **kwargs) -> Self:
        raise NotImplementedError(f"")

    def update(self, value, **kwargs) -> Self:
        raise NotImplementedError(f"")

    def remove(self, **kwargs) -> int:
        raise NotImplementedError(f"")

    def find(self, *args, default_value=_not_found_, **kwargs) -> typing.Any:
        request = self._mapper.child(self._path).find(*args, default_value=_not_found_, lazy=False, **kwargs)

        return self._op_find(request, default_value=default_value)

    def for_each(self, *args, **kwargs) -> typing.Generator[typing.Tuple[int, typing.Any], None, None]:
        """Return a generator of the results."""
        for idx, request in self._mapper.child(self._path).for_each(*args, **kwargs):
            yield idx, self._op_find(request)

    def search(self, *args, **kwargs):
        raise NotImplementedError()

    def _get_entry_by_name(self, entry_name: str, default_value=None) -> Entry | None:
        entry = self._entry_list.get(entry_name, None)

        if isinstance(entry, (str, URITuple)):
            entry = open_entry(entry)
            self._entry_list[entry_name] = entry

        if isinstance(entry, Entry):
            pass
        elif default_value is _not_found_:
            raise RuntimeError(f"Can not find entry for {entry_name}")
        else:
            entry = default_value

        return entry

    def _op_find(self, request: typing.Any, *args, **kwargs) -> typing.Any:
        if isinstance(request, str) and "://" in request:
            request = uri_split_as_dict(request)

        if request is _not_found_ or request is None:
            defaultentry = self._get_entry_by_name("*", None)
            if defaultentry is None:
                res = _not_found_
            else:
                res = defaultentry.child(self._path).find(None, *args, **kwargs)

        elif isinstance(request, Entry):
            res = EntryProxy(request, self._entry_list)

        elif isinstance(request, list):
            res = [self._op_find(req, *args, **kwargs) for req in request]

        elif not isinstance(request, dict):
            res = request

        elif "@spdb" not in request:
            res = {k: self._op_find(req, *args, **kwargs) for k, req in request.items()}

        else:
            entry = self._get_entry_by_name(request.get("@spdb", None))

            if not isinstance(entry, Entry):
                raise RuntimeError(f"Can not find entry for {request}")

            res = entry.find(request.get("_text"), *args, **kwargs)

        return res



from __future__ import annotations

import collections.abc
import typing
import numpy as np
from copy import deepcopy
from .Entry import Entry
from .HTree import List, HTree
from .Path import update_tree, merge_tree, Path
from .sp_property import SpTree, sp_property
from ..utils.tags import _not_found_
from ..utils.logger import logger


class TimeSlice(SpTree):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iteration = 0

    time: float = sp_property(unit="s", default_value=0.0)  # type: ignore

    @property
    def iteration(self) -> int:
        return self._iteration

    def refresh(self, value=_not_found_, *args, **kwargs):
        current = self._update_(None, value, *args, **kwargs)
        current._iteration += 1
        return current


_TSlice = typing.TypeVar("_TSlice", bound=TimeSlice)


class TimeSeriesAoS(List[_TSlice]):
    """A series of time slices .

    用以管理随时间变化（time series）的一组状态（TimeSlice）。

    current:
        指向当前时间片，即为序列最后一个时间片吗。

    _TODO_
      1. 缓存时间片，避免重复创建，减少内存占用
      2. 缓存应循环使用
      3. cache 数据自动写入 entry 落盘
    """

    def __init__(self, *args, cache_depth=3, **kwargs):
        super().__init__(*args, **kwargs)

        if self._cache is _not_found_ or self._cache is None:
            self._cache = []

        self._entry_cursor = None
        self._cache_cursor = len(self._cache) - 1
        self._cache_depth = cache_depth

        if self._cache_cursor + 1 < self._cache_depth:
            self._cache += [_not_found_] * (self._cache_depth - self._cache_cursor - 1)
        else:
            self._cache_depth = self._cache_cursor + 1

    def dump(self, entry: Entry, **kwargs) -> None:
        """将数据写入 entry"""
        entry.insert([{}] * len(self._cache))
        for idx, value in enumerate(self._cache):
            if isinstance(value, HTree):
                value.dump(entry.child(idx), **kwargs)
            else:
                entry.child(idx).insert(value)

    def __missing__(self, idx: int):
        """当循环队列满了或序号出界的时候调用
        :param o: 最老的 time_slice
        """
        return super().__missing__(idx)

    @property
    def time(self) -> float:
        return self.current.time

    @property
    def iteration(self) -> int:
        return self.current.iteration

    @property
    def dt(self) -> float:
        return self._metadata.get("dt", 0.1)

    @property
    def current(self) -> _TSlice:
        return self._find_(0)

    @property
    def previous(self) -> typing.Generator[_TSlice, None, None]:
        for i in range(len(self._cache)):
            yield self._find_(-(i + 1))

    @property
    def is_initializied(self) -> bool:
        return self._cache_cursor >= 0

    def _find_slice_by_time(self, time: float) -> typing.Tuple[int, float]:
        if self._entry is None:
            return None, None

        time_coord = getattr(self, "_time_coord", _not_found_)

        if time_coord is _not_found_:
            time_coord = self._metadata.get("coordinate1", _not_found_)

        if isinstance(time_coord, str):
            time_coord = self.get(time_coord, default_value=_not_found_)
            if time_coord is None:
                time_coord = self._entry.child(time_coord).find()

        self._time_coord = time_coord

        pos = None

        if isinstance(time_coord, np.ndarray):
            indices = np.where(time_coord < time)[0]
            if len(indices) > 0:
                pos = indices[-1] + 1
                time = time_coord[pos]

        elif self._entry is not None:
            pos = self._entry_cursor or 0

            while True:
                t_time = self._entry.child(f"{pos}/time").find(default_value=_not_found_)

                if t_time is _not_found_ or t_time is None or t_time > time:
                    time = None
                    break
                elif np.isclose(t_time, time):
                    time = t_time
                    break
                else:
                    pos = pos + 1

        return pos, time

    def _find_(self, idx: int, *args, **kwargs) -> _TSlice:
        if not isinstance(idx, int):
            return _not_found_
        elif not self.is_initializied:
            self.initialize(*args, **kwargs)

        cache_pos = (self._cache_cursor + idx + self._cache_depth) % self._cache_depth

        value = self._cache[cache_pos]

        if not (value is _not_found_ or isinstance(value, TimeSlice)):
            if isinstance(self._entry, Entry) and self._entry_cursor is not None:
                entry_cursor = self._entry_cursor + idx
                entry = self._entry.child(entry_cursor)
            else:
                entry_cursor = 0
                entry = None
            value = self._type_convert(value, cache_pos, _entry=entry, _parent=self._parent)

        return value

    def initialize(self, *args, **kwargs):
        """初始化 TimeSeries"""
        if self.is_initializied:
            return
            # raise RuntimeError(f"TimeSeries is already initialized!")

        self._cache_cursor = 0

        self._cache[self._cache_cursor] = update_tree(self._cache[self._cache_cursor], *args, kwargs)

        current = self._cache[self._cache_cursor]

        time = Path("time").get(current, 0.0)

        self._entry_cursor, time_hint = self._find_slice_by_time(time)

        default_value = deepcopy(self._metadata.get("default_initial_value", {}))

        if time_hint is not None:
            time = time_hint

        current = update_tree(default_value, current, {"time": time})

        self._cache[self._cache_cursor] = current

    def refresh(self, *args, **kwargs) -> typing.Type[TimeSeriesAoS]:
        if not self.is_initializied:
            raise RuntimeError("TimeSlice is not initializied!")

        self._cache[self._cache_cursor] = update_tree(self._cache[self._cache_cursor], *args, **kwargs)

    def advance(self, *args, **kwargs) -> _TSlice:
        if not self.is_initializied:
            raise RuntimeError("TimeSlice is not initializied!")

        self._cache_cursor = (self._cache_cursor + 1) % self._cache_depth

        if self._cache[self._cache_cursor] is not _not_found_:
            self.__full__(self._cache[self._cache_cursor])

        self._entry_cursor += 1

        self._cache[self._cache_cursor] = merge_tree(*args, **kwargs)

        return self.current

    def finalize(self):
        """完成"""
        pass



from .Mesh import Mesh
from ..utils.typing import ArrayType
from .mesh_rectilinear import RectilinearMesh


@Mesh.register(["rectangular", "rect"])
class RectangularMesh(RectilinearMesh):
    """ Rectangular Mesh, which is alias of RectilinearMesh
    矩形网格
    """

    @property
    def dim1(self) -> ArrayType: return self._dims[0]

    @property
    def dim2(self) -> ArrayType: return self._dims[1]






import numpy as np
from scipy import constants
from spdm.utils.logger import logger
from spdm.data.Function import Piecewise, Function


if __name__ == '__main__':
    x1 = np.linspace(-1, 1, 128)
    x2 = np.linspace(-1, 1, 64)
    fun1 = Function(np.sin(x1), x1)
    fun2 = Function(np.cos(x2), x2)

    expr = fun1 + fun2

    logger.debug(expr)

    logger.debug(expr(x2))

    fun3 = Piecewise([np.sin, np.cos], [lambda x:x > 0, lambda x:x < 0])
    logger.debug(fun3(x1))



import pprint
import matplotlib.pyplot as plt
import numpy as np
import pathlib
from spdm.core.File import File
from ..util.logger import logger
import sys


if __name__ == '__main__':
    doc = File(pathlib.Path(__file__).parent/"data/g063982.04800", format="GEQdsk")
    pprint.pprint(doc.root._holder)



from __future__ import annotations
import collections.abc
import functools
import typing
from copy import deepcopy

from typing_extensions import Self

from spdm.core.HTree import HTree, HTreeNode
from spdm.utils.tags import _not_found_

from .Entry import Entry
from .HTree import HTree, List, Dict
from .Path import Path, PathLike, as_path, OpTags, update_tree, merge_tree
from ..utils.tags import _not_found_, _undefined_
from ..utils.typing import array_type, get_args, get_type_hint
from ..utils.logger import logger

_T = typing.TypeVar("_T")


class QueryResult(HTree):
    """Handle the result of query"""

    def __init__(self, query: PathLike, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._path = as_path(query)

    def __getattr__(self, name: str):
        return self._get(name)

    def _get(self, query: str | int | slice | dict, *args, **kwargs):
        default_value = kwargs.pop("default_value", _not_found_)
        _VT = get_args(self.__orig_class__)[0]
        if isinstance(query, str):
            if default_value is _not_found_ and isinstance(self._default_value, dict):
                default_value = self._default_value.get(query, _not_found_)
            tp = get_type_hint(_VT, query)

            return QueryResult[tp](self._path.append(query), *args, default_value=default_value, **kwargs)
        else:
            return QueryResult[_VT](self._path.append(query), *args, default_value=default_value, **kwargs)

    @property
    def _value_(self) -> typing.Any:
        value = super()._query(self._path)
        if isinstance(value, list):
            value = functools.reduce(self._default_reducer, value)
        return value

    def __call__(self, *args, **kwargs) -> typing.Any:
        value = super()._query(self._path, op=Path.tags.call, *args, **kwargs)

        if isinstance(value, list):
            value = functools.reduce(self._default_reducer, value)

        return value

    @staticmethod
    def _default_reducer(first: typing.Any, second: typing.Any) -> typing.Any:
        if first is _not_found_:
            return second
        elif second is _not_found_ or second is None:
            return second
        elif isinstance(first, (str)):
            return first
        elif isinstance(first, array_type) and isinstance(second, array_type):
            return first + second
        elif isinstance(first, (dict, list)) or isinstance(second, (dict, list)):
            return update_tree(first, second)
        else:
            return first + second

    def children(self) -> typing.Generator[_T | HTree, None, None]:
        """遍历 children"""
        cache = self._cache if self._cache is not _not_found_ else self._default_value

        if not isinstance(cache, list) or len(cache) == 0:
            yield from super().children()

        else:
            for idx, value in enumerate(cache):
                if isinstance(value, (dict, Dict)):
                    id = value.get(self._identifier, None)
                else:
                    id = None
                if id is not None:
                    entry = self._entry.child({f"@{self._identifier}": id})
                else:
                    entry = None

                yield self._type_convert(value, idx, entry=entry)


_TTree = typing.TypeVar("_TTree")


class AoS(List[_TTree]):
    """
    Array of structure

    FIXME: 需要优化！！
        - 数据结构应为 named list or ordered dict
        - 可以自动转换 list 类型 cache 和 entry
    """

    def __missing__(self, key) -> _TTree:
        tag = f"@{Path.id_tag_name}"

        if self._cache is not None and len(self._cache) > 0:
            pass
        elif self._entry is not None:
            keys = set([key for key in self._entry.child(f"*/{tag}").for_each()])
            self._cache = [{tag: key} for key in keys]
        else:
            self._cache = []

        value = deepcopy(self._metadata.get("default_initial_value", _not_found_) or {})

        value[f"@{Path.id_tag_name}"] = key

        # self._cache.append(value)

        return value

    def _find_(self, key, *args, default_value=_undefined_, **kwargs) -> _TTree:
        """AoS._find_ 当键值不存在时，默认强制调用 __missing__"""
        self._sync_cache()
        if default_value is _not_found_:
            default_value = _undefined_

        return super()._find_(key, *args, default_value=default_value, **kwargs)

    def _update_(self, key, value, op=None, *args, **kwargs) -> Self:
        if (key is None or key is _not_found_) and op is not Path.tags.extend:
            key = as_path(f"@{Path.id_tag_name}").get(value, None)

            if key is _not_found_ or key is None:
                raise KeyError(f"key is not found in {value} {op}")

        super()._update_(key, value, op, *args, **kwargs)

        return self

    def _sync_cache(self):
        if len(self._cache) == 0 and self._entry is not None:
            keys = self._entry.child(f"*/@{Path.id_tag_name}").get()
            if isinstance(keys, collections.abc.Sequence):
                keys = set([k for k in keys if k is not _not_found_ and k is not None])
                self._cache = [{f"@{Path.id_tag_name}": k} for k in keys]

    def _for_each_(self, *args, **kwargs) -> typing.Generator[typing.Tuple[int | str, HTreeNode], None, None]:
        self._sync_cache()
        tag = f"@{Path.id_tag_name}"
        if len(self._cache) > 0:
            for idx, v in enumerate(self._cache):
                if isinstance(v, (dict, HTree)) and (key := Path(tag).get(v, _not_found_)) is not _not_found_:
                    yield idx, self._find_(key, *args, **kwargs)
                else:
                    yield idx, self._find_(idx, *args, **kwargs)
        elif self._entry is not None:
            for idx, e in self._entry.for_each():
                if isinstance(e, Entry):
                    yield idx, self._type_convert(_not_found_, idx, _entry=e)
                else:
                    yield idx, self._type_convert(e, idx, _entry=self._entry.child(idx))

        # if self._entry is None:
        #     _entry = None
        # elif key is _not_found_ or v is None:
        #     _entry = self._entry.child(idx)
        # else:
        #     _entry = self._entry.child({tag: key})
        # yield self._type_convert(v, idx, _entry=_entry)

    def fetch(self, *args, _parent=_not_found_, **kwargs) -> Self:
        return self.__duplicate__([HTreeNode._do_fetch(obj, *args, **kwargs) for obj in self], _parent=_parent)

    def dump(self, entry: Entry, **kwargs) -> None:
        """将数据写入 entry"""
        entry.insert([{}] * len(self._cache))
        for idx, value in enumerate(self._cache):
            if isinstance(value, HTree):
                value.dump(entry.child(idx), **kwargs)
            else:
                entry.child(idx).insert(value)



import collections
import contextlib
from copy import copy

import jsonschema

from . import io
from .Alias import Alias
from .logger import logger
from .uri_utils import getvalue_r, uri_join


def _extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def _properties(validator, properties, instance, schema):
        for p, subschema in properties.items():
            if isinstance(subschema, collections.abc.Iterable) \
                    and "default" in subschema \
                    and hasattr(instance, "setdefault"):
                instance.setdefault(p, subschema["default"])

        for error in validate_properties(validator, properties, instance, schema,):
            yield error

    return jsonschema.validators.extend(
        validator_class, {"properties": _properties},
    )


_DefaultValidatingValidator = _extend_with_default(jsonschema.Draft7Validator)


class RefResolver(object):
    """ Resolve and fetch '$ref' in the document (json,yaml,https ...)
        schemas:
            pkgdata     :   pkgutil.get_data(o.authority or __package__,o.path)
            https,http  :   requests.get(uri).json()
            josn        :   json.load(open(uri))
            yaml        :   json.load(open(uri))
        Example(code):

            >>> repo = RefResolver()
            >>> repo.alias.append("https://fusionyun.org/schemas/draft-00/",
                        "pkgdata:///../schemas")
            >>> repo.alias.append(f"/modules/", "file://workspaces/FyPackages/modules/")
        @note
            *  compatible with jsonschema.RefResolver

        TODO (salmon 20190915): support XML,XSD,XSLT
    """

    def __init__(self, *,
                 base_uri="",
                 encode='UTF-8',
                 prefetch=None,
                 enable_remote=False,
                 enable_validate=True,
                 enable_envs_template=True,
                 default_file_ext='yaml',
                 default_schema="http://json-schema.org/draft-07/schema#",
                 alias=None,
                 **kwargs
                 ):
        super().__init__()

        self._alias = Alias(glob_pattern_char='*')
        self._encode = encode
        self._scopes_stack = [base_uri] if len(
            base_uri) > 0 and base_uri[-1] == '/' else [base_uri+"/"]
        self._default_file_ext = default_file_ext
        self._default_schema = uri_join(base_uri, default_schema)
        self._enable_remote = enable_remote
        self._enable_validate = enable_validate
        self._enable_envs_template = enable_envs_template
        if prefetch is not None:
            # if not isinstance(prefetch, pathlib.Path):
            #     prefetch = pathlib.Path(prefetch)

            # if prefetch.is_dir():
            prefetch = f"{prefetch}/" if prefetch[-1] != '/' else prefetch
            self._alias.prepend("https://", prefetch)
            self._alias.prepend("http://", prefetch)

        if alias is None:
            pass
        elif isinstance(alias, collections.abc.Mapping):
            for k, v in alias.items():
                self._alias.append(self.normalize_uri(k), v)
        elif isinstance(alias, collections.abc.Sequence):
            for k, v in alias:
                self._alias.append(self.normalize_uri(k), v)
        else:
            raise TypeError(f"Require list or map, not [{type(alias)}]")

        self._cache = {}
        self._validator = {"http://json-schema.org/draft-07/schema#":
                           _DefaultValidatingValidator}

    @property
    def alias(self):
        return self._alias

    def normalize_uri(self, uri):
        if uri is None:
            uri = None
        elif type(uri) is str:
            pass
        elif isinstance(uri, collections.abc.Sequence):
            uri = "/".join(uri)
        else:
            raise TypeError(f"Illegal type {type(uri).__name__}")
        return uri_join(self.resolution_scope, uri)

    def remove_prefix(self, p: str):
        return self.relative_path(p)

    def relative_path(self, p: str):
        # FIXME: not complete , only support sub-path
        prefix = self.resolution_scope
        if p is None:
            return None
        elif p.startswith(prefix):
            return p[len(prefix):].strip("/.")
        else:
            raise NotImplementedError()

    _normalize_ids = ["$schema", "$id", "$base"]

    def validate(self, doc):
        if doc is None:
            raise ValueError(f"Try to validate an empty document!")

        for nid in RefResolver._normalize_ids:
            if nid not in doc:
                continue
            doc[nid] = self.normalize_uri(doc[nid])

        schema = doc.get("$schema", None)
        if isinstance(schema, str):
            schema_id = schema
            schema = {"$id": schema_id}
        elif isinstance(schema, collections.abc.Mapping):
            schema_id = schema.get("$id", None)
        else:
            schema_id = None
            schema = {"$id": schema_id}

        validator = self._validator.get(schema_id, None)
        if validator is None:
            try:
                schema = self.fetch(schema, no_validate=True)
            except Exception:
                logger.error(f"Can not find schema : {schema}")
            else:
                validator = _DefaultValidatingValidator(schema, resolver=self)
                self._validator[schema["$id"]] = validator

        if validator is not None:
            validator.validate(doc)

        return doc

    def _do_fetch(self, uri):
        uri = self.normalize_uri(uri)
        new_doc = self._cache.get(uri, None)
        if new_doc is not None:
            return new_doc

        for a_uri in self._alias.match(uri):
            new_doc = io.read(a_uri)

            if not new_doc:
                pass
            else:
                new_doc["$id"] = uri
                new_doc["$source_file"] = a_uri
                self._cache[uri] = new_doc
                break

        return new_doc

    def fetch(self, doc, no_validate=False) -> dict:
        """ fetch document from source, then validate and fill default
            value basing on $schema in document
        """

        if isinstance(doc, (str, collections.abc.Sequence)):
            new_doc = self._do_fetch(doc)
        elif isinstance(doc, collections.abc.Mapping):
            new_doc = copy(doc)
        else:
            raise TypeError(type(doc))

        if isinstance(new_doc, collections.abc.Mapping):
            # new_doc["$schema"] = self.normalize_uri(new_doc.get("$schema", ""))
            if not no_validate and self._enable_validate:
                self.validate(new_doc)

        return new_doc

    def clear_cache(self):
        self._cache.clear()  # pylint: disable= no-member

    def glob(self, mod=None):
        mod_prefix = self.normalize_uri(f"{mod or ''}%_PATH_%")

        for n_uri in self._alias.match(mod_prefix):
            for p, f in io.glob(n_uri):
                yield p, f

    ##########################################################################
    # Begin: compatible with jsonschema.RefResolver

    @classmethod
    def from_schema(cls_, schema, base_uri=None):
        # if schema is None:
        #     return cls_()
        # s_id = schema.get("$id", "")
        # base_uri = base_uri or cls_._base_uri
        # res = cls_(base_uri=base_uri)
        # res.insert(s_id, schema)
        return cls_()

    def push_scope(self, scope):
        self._scopes_stack.append(uri_join(self.resolution_scope, scope))

    def pop_scope(self):
        try:
            self._scopes_stack.pop()
        except IndexError:
            raise IndexError("Failed to pop from an empty stack")

    @property
    def resolution_scope(self):
        return self._scopes_stack[-1]

    @contextlib.contextmanager
    def in_scope(self, scope):
        self.push_scope(scope)
        try:
            yield
        finally:
            self.pop_scope()

    @contextlib.contextmanager
    def resolving(self, ref):
        uri, resolved = self.resolve(ref)
        self.push_scope(uri)
        try:
            yield resolved
        finally:
            self.pop_scope()

    def resolve(self, ref):
        """ Parse reference or description, return URI and full schema"""
        uri = self.normalize_uri(ref)
        return uri, self.fetch(uri, no_validate=True)

    def resolve_from_uri(self, uri):
        return self.fetch(uri, no_validate=True)

    def resolve_remote(self, uri):
        return self.fetch(uri, no_validate=True)

    def resolve_local(self, local_path):
        return self.fetch(uri_join("local://", local_path), no_validate=True)

    def resolve_fragment(self, obj, fragment):
        return getvalue_r(obj, fragment)

    # END :compatible with jsonschema.RefResolver
    ##########################################################################

        # RefResolver.HANDLERS["pyobject"] = lambda p: {
        #     "$schema": "PyObject", "$class": p}



from __future__ import annotations
import collections
import collections.abc
import inspect
import abc
import typing
import math
import dataclasses
from typing_extensions import Self
from copy import deepcopy
from .Path import Path, as_path
from .HTree import HTree, HTreeNode, Dict
from .sp_property import PropertyTree, SpTree
from ..utils.logger import logger
from ..utils.tags import _not_found_, _undefined_
from ..utils.typing import array_type, isinstance_generic


class Port:
    def __init__(self, node=_not_found_, path: str | Path = None, type_hint: typing.Type = None, **kwargs) -> None:
        if isinstance(node, Port):
            type_hint = type_hint or node.type_hint
            node = node.node

        self.node: HTreeNode = _not_found_

        path = as_path(path)

        self.identifier = path[0]

        self.fragment = as_path(path[1:])

        self.type_hint: typing.Type = type_hint

        self._metadata = kwargs

        self.link(node)

    def link(self, node=_not_found_, type_hint: typing.Type = _not_found_):
        if type_hint is not _not_found_:
            self.type_hint = type_hint

        if node is _not_found_:
            pass

        elif isinstance(node, Port):
            self.link(node.node)

        elif node is not _not_found_ and node is not None and node is not _undefined_:
            self.node = node

        return self.node

    def unlink(self):
        self.node = _not_found_
        self.fragment = Path()

    def __copy__(self):
        other = Port()
        other._metadata = deepcopy(self._metadata)
        other.link(self.node, fragment=self.fragment, type_hint=self.type_hint)
        return other

    def __getitem__(self, key):
        return self.fragment.append(key).find(self.node)

    def __setitem__(self, key, value):
        return self.fragment.append(key).update(self.node, value)

    def insert(self, *args, **kwargs):
        return self.fragment.insert(self.node, *args, **kwargs)

    def update(self, *args, **kwargs):
        if (self.node is _not_found_ or self.node is None) and len(self.fragment) == 0:
            self.node = args[0]
            args = args[1:]

        if len(args) + len(kwargs) > 0:
            self.fragment.update(self.node, *args, **kwargs)

    def remove(self, *args, **kwargs):
        return self.fragment.remove(self.node, *args, **kwargs)

    def find(self, *args, **kwargs):
        return self.fragment.find(self.node, *args, **kwargs)

    def fetch(self, *args, **kwargs):
        node = self.fragment.find(self.node)
        return HTreeNode._do_fetch(node, *args, **kwargs)

    @property
    def is_changed(self) -> bool:
        return not (
            math.isclose(getattr(self.node, "time", 0), self._time)
            and (getattr(self.node, "iteration", None) == self._iteration)
        )


class Ports(Dict[Port]):
    """Port 的汇总，

    Args:
        typing (_type_): _description_
    """

    def get(self, key: str, *args, **kwargs) -> Port:
        port: Port = super().find(key, *args, **kwargs)
        if (port.node is _not_found_ or port.node is None) and len(port.fragment) > 0:
            port0 = super().get(port.identifier, _not_found_)
            if isinstance(port0, Port):
                port.node = port0.node
        return port

    def put(self, key: str, value, *args, **kargs) -> None:
        return self.get(key).update(value)

    def __missing__(self, key: str) -> Port:
        self._cache[key] = port = Port(path=key)
        return port

    def refresh(self, *args, **kwargs) -> Self:
        attr_name = self.__class__.__name__.lower()

        if len(args) + len(kwargs) > 0:
            for obj in [*args, kwargs]:
                if isinstance(obj, Ports):
                    for k, n in self.items():
                        n.link(obj.get_cache(k, _not_found_) or obj.get_cache(n.identifier, _not_found_))
                elif isinstance(obj, collections.abc.Mapping):
                    for k, n in self.items():
                        n.link(obj.get(k, obj.get(n.identifier, _not_found_)))

        else:
            parent: HTreeNode = as_path("../../").get(self, _not_found_)
            if isinstance(parent, collections.abc.Sequence):
                parent = getattr(parent, "_parent", _not_found_)

            if isinstance(parent, SpTree):
                for n in self.values():
                    node = getattr(parent, n.identifier, _not_found_)
                    n.link(node)

            self.refresh(getattr(parent, attr_name, _not_found_))

        return self


class InPorts(Ports):
    def edge(self, name, *args, **kwargs) -> Edge:
        return Edge(self[name], self._holder, *args, **kwargs)


class OutPorts(Ports):
    def edge(self, name, *args, **kwargs) -> Edge:
        return Edge(self._holder, self[name], *args, **kwargs)


class Edge:
    """`Edge` defines a connection between two `Port`s

    Attribute

    - source      : the start of edge which must be `OUTPUT Port`
    - target      : the start of edge which must be `INPUT Port`
    - dtype       : defines what `Port`s it can be connected, (default: string)
    - label       : short string
    - description : long string
    """

    def __init__(
        self,
        source=None,
        target=None,
        source_type_hint=None,
        target_type_hint=None,
        graph=None,
        **kwargs,
    ):
        self._source = Port(source, source_type_hint)

        self._target = Port(target, target_type_hint)

        self._graph = graph

        self._metadata = PropertyTree(kwargs)

    def __copy__(self):
        return Edge(self._source, self._target, graph=self._graph, **self._metadata._cache)

    @property
    def metadata(self) -> PropertyTree:
        return self._metadata

    @property
    def source(self) -> Port:
        return self._source

    @property
    def target(self) -> Port:
        return self._target

    @property
    def is_linked(self):
        return self._target.node is not None and self._source.node is not None

    def __str__(self):
        return f"""<{self.__class__.__name__} source='{self._source}' target='{self._target}' label='{self._metadata.get('label','')}'/>"""

    def _repr_s(self):
        def _str(s):
            if s is None:
                return ""
            elif type(s) is str:
                return "." + s
            elif isinstance(s, slice):
                if s.stop is None:
                    return f"[{s.start or ''}:{s.step or ''}]"
                else:
                    return f"[{s.start or ''}:{s.step or ''}:{s.stop or ''}]"

            elif isinstance(s, collections.abc.Sequence):
                return "".join([f"{_str(t)}" for t in s])
            else:
                return f"[{s}]"

        return "".join([_str(s) for s in self._path])

    def split(self, *args, **kwargs):
        """
        using Slot Node split edge into chain, add In(Out)Slot not to graph

        return list of splitted edges
        """
        source = self._source
        target = self._target

        s = collections.deque([source._parent] if source is not None else [])
        t = collections.deque([target._parent] if target is not None else [])

        if getattr(s[0], "parent", None) is getattr(t[0], "parent", None):
            return self

        while s[0] is not None:
            s.appendleft(s[0]._parent)

        while t[0] is not None:
            t.appendleft(t[0]._parent)

        s_rank = len(s)
        t_rank = len(t)
        pos = s_rank - 2
        tag = ""
        while pos >= t_rank or (pos >= 0 and s[pos] is not t[pos]):
            tag = f"{tag}_{source._parent.name}"

            s[pos].slot[tag] = source

            source = s[pos].port[tag]

            pos = pos - 1

        tag = f"{tag}_{source._parent.name}"

        while pos < t_rank - 2:
            pos = pos + 1
            tag = f"{tag}_{source._parent.parent.name}"
            t[pos].port[tag] = source

            source = t[pos].slot[tag]

        src, prefix = self._unwrap(source)
        self._source = src
        self._path = prefix + self._path
        return self



import typing

from ..utils.typing import ArrayType
from .GeoObject import GeoObject
from .Line import Line
from .Plane import Plane
from .Point import Point
from .Solid import Solid
from .Surface import Surface
from .BBox import BBox


@GeoObject.register("circle")
class Circle(GeoObject):
    """ Circle
        圆，具有一个固定圆心和一个固定半径
    """

    def __init__(self, x: float, y: float, r: float, **kwargs) -> None:
        super().__init__(rank=1, ndims=2, is_closed=True, **kwargs)
        self._x = x
        self._y = y
        self._r = r

    @property
    def bbox(self) -> BBox:
        return BBox([self._x - self._r, self._y - self._r], [2.0*self._r, 2.0 * self._r])

    @property
    def x(self) -> float: return self._x

    @property
    def y(self) -> float: return self._y

    @property
    def r(self) -> float: return self._r

    def map(self, u, *args, **kwargs):
        return NotImplemented

    def derivative(self, u, *args, **kwargs):
        return NotImplemented

    def dl(self, u, *args, **kwargs):
        return NotImplemented

    def pullback(self, func, *args, **kwargs):
        return NotImplemented

    def make_one_form(self, func):
        return NotImplemented


@GeoObject.register("ellipse")
class Ellipse(GeoObject):
    """ Ellipse
        椭圆，具有一个固定圆心和两个固定半径
    """
    pass


@Plane.register("disc")
class Disc(Plane):
    """ Disc
        圆盘
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._boundary = Circle(*args)

    @property
    def boundary(self) -> Circle:
        return self._boundary


@Surface.register("sphere")
class Sphere(Surface):
    """ Sphere
        球面
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._boundary = Circle(*args)

    pass


@Solid.register("ball")
class Ball(GeoObject):
    """ Ball
        球体
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._boundary = Sphere(*args)


@Solid.register("cylinder")
class Cylinder(GeoObject):
    """ Cylinder
        圆柱体，具有两个固定端面
    """
    pass


@Surface.register("toroidal_surface")
class ToroidalSurface(Surface):
    def __init__(self, cross_section: Line, circle: Circle, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


@Solid.register("toroidal")
class Toroidal(Solid):
    def __init__(self, section: Plane, circle: Circle, *args, **kwargs) -> None:
        super().__init__(**kwargs)
        self._boundary = ToroidalSurface(cross_section.boundary, circle, *args,)



import collections.abc
import typing
from functools import cached_property

import numpy as np
from scipy import interpolate

from ..geometry.Curve import Curve
from ..geometry.GeoObject import GeoObject, GeoObjectSet
from ..geometry.Point import Point
from ..geometry.Surface import Surface
from ..utils.logger import logger
from ..utils.typing import ArrayType, ScalarType
from .Mesh import Mesh
from .mesh_rectilinear import RectilinearMesh


@Mesh.register("curvilinear")
class CurvilinearMesh(RectilinearMesh):
    """A `curvilinear Mesh` or `structured Mesh` is a Mesh with the same combinatorial structure as a regular Mesh,
in which the cells are quadrilaterals or [general] cuboids, rather than rectangles or rectangular cuboids.
-- [https://en.wikipedia.org/wiki/Regular_Mesh]
"""
    TOLERANCE = 1.0e-5

    def __init__(self, *args, ** kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert isinstance(self._geometry, GeoObjectSet)
        assert isinstance(self._dims, collections.abc.Sequence)

    def axis(self, idx, axis=0):
        if axis == 0:
            return self._geometry[idx]
        else:
            s = [slice(None, None, None)]*self.ndims
            s[axis] = idx
            s = s+[slice(None, None, None)]

            sub_xy = self.xy[tuple(s)]  # [p[tuple(s)] for p in self._xy]
            sub_uv = [self._uv[(axis+i) % self.geometry.ndim]
                      for i in range(1, self.geometry.ndim)]
            sub_cycles = [self.cycles[(axis+i) % self.geometry.ndim]
                          for i in range(1, self.geometry.ndim)]

            return CurvilinearMesh(sub_xy, sub_uv,  cycles=sub_cycles)

    @property
    def uv(self) -> ArrayType: return self._uv

    @cached_property
    def points(self) -> typing.List[ArrayType]:
        if isinstance(self.geometry, GeoObject):
            return self.geometry.points()
        elif isinstance(self.geometry, GeoObjectSet):
            d = np.stack([surf.points for surf in self.geometry], axis=0)
            return tuple(d[..., i] for i in range(self.geometry.ndims))
        else:
            raise RuntimeError(f'Unknown type {type(self.geometry)}')

    @cached_property
    def volume_element(self) -> ArrayType:
        raise NotImplementedError()

    @property
    def xyz(self): return self.points

    # def pushforward(self, new_uv):
    #     new_shape = [len(u) for u in new_uv]
    #     if new_shape != self.shape:
    #         raise ValueError(f"illegal shape! {new_shape}!={self.shape}")
    #     return CurvilinearMesh(self._xy, new_uv, cycles=self.cycles)

    def interpolator(self, value,  **kwargs):
        if value.shape != self.shape:
            raise ValueError(f"{value.shape} {self.shape}")

        if self.ndims == 1:
            interp = interpolate.InterpolatedUnivariateSpline(self._dims[0], value,  **kwargs)
        elif self.ndims == 2:
            interp = interpolate.RectBivariateSpline(self._dims[0], self._dims[1], value, ** kwargs)
        else:
            raise NotImplementedError(f"NDIMS {self.ndims}>2")
        return interp

    @ cached_property
    def boundary(self):
        return {"inner": self.axis(0, 0),  "outer": self.axis(-1, 0)}

    @ cached_property
    def geo_object(self):
        if self.rank == 1:
            if all([np.var(x)/np.mean(x**2) < CurvilinearMesh.TOLERANCE for x in self.xy.T]):
                gobj = Point(*[x[0] for x in self.xy.T])
            else:
                gobj = CubicSplineCurve(self.xy, self._uv[0], is_closed=self.cycles[0])
        elif self.rank == 2:
            gobj = BSplineSurface(self.xy, self._uv,  is_closed=self.cycles)
        else:
            raise NotImplementedError()

        return gobj



import os

SP_DEBUG = os.environ.get("SP_DEBUG", True)

SP_LABEL = os.environ.get("SP_LABEL", __package__[: __package__.find(".")])

SP_MPI = None
SP_MPI_RANK = 0
SP_MPI_SIZE = 0
if int(os.environ.get("OMPI_COMM_WORLD_SIZE", "0")) > 0:
    try:
        from mpi4py import MPI as SP_MPI

        SP_MPI_RANK = SP_MPI.COMM_WORLD.Get_rank()
        SP_MPI_SIZE = SP_MPI.COMM_WORLD.Get_size()
    except ImportError:
        pass
try:
    from ..__version__ import __version__
except ImportError:
    SP_VERSION = "beta"
else:
    SP_VERSION = __version__


def is_notebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


SP_QUIET = os.environ.get("SP_QUIET", is_notebook())



import pprint
import typing
import unittest
from copy import deepcopy

from spdm.core.HTree import Dict, HTree, List, HTreeNode
from spdm.utils.logger import logger
from spdm.utils.tags import _undefined_
from spdm.utils.typing import as_value


class Foo(Dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


test_data = {
    "a": ["hello world {name}!", "hello world2 {name}!", 1.0, 2, 3, 4],
    "c": "I'm {age}!",
    "d": {"e": "{name} is {age}", "f": "{address}"},
}


class NamedFoo(Dict):
    a: List[typing.Any]
    c: str
    d: Dict


class TestHTree(unittest.TestCase):
    def test_empty_node(self):
        n = HTreeNode()

        self.assertTrue(n.__empty__())
        self.assertFalse(n.__null__())
        self.assertEqual(len(n) == 0)

    def test_null_node(self):
        n = HTreeNode(None)

        self.assertFalse(n.__empty__())
        self.assertTrue(n.__null__())
        self.assertEqual(len(n) == 0)

    data = {
        "a": ["hello world {name}!", "hello world2 {name}!", 1, 2, 3, 4],
        "c": "I'm {age}!",
        "d": {"e": "{name} is {age}", "f": "{address}"},
    }

    # def test_new(self):
    #     self.assertTrue(isinstance(HTree("hello"), HTree))
    #     self.assertTrue(isinstance(HTree(1), HTree))
    #     self.assertTrue(isinstance(HTree(np.ones([10, 20])), HTree))
    #     self.assertTrue(isinstance(HTree([1, 2, 3, 4, 5]), List))
    #     self.assertTrue(isinstance(HTree((1, 2, 3, 4, 5)), List))
    #     self.assertTrue(isinstance(HTree({"a": 1, "b": 2, "c": 3}), Dict))
    #     # self.assertFalse(isinstance(HTree({1, 2, 3, 4, 5}), List))

    # def test_create(self):
    #     cache = []
    #     d = HTree(cache)
    #     self.assertEqual(d.create_child("hello"), "hello")
    #     self.assertEqual(d.create_child(1), 1)
    #     v = np.ones([10, 20])
    #     self.assertIs(d.create_child(v), v)
    #     self.assertTrue(isinstance(d.create_child("hello", always_node=True), Node))

    #     self.assertTrue(isinstance(d.create_child([1, 2, 3, 4, 5]), List))
    #     self.assertTrue(isinstance(d.create_child((1, 2, 3, 4, 5)), List))
    #     self.assertTrue(isinstance(d.create_child({"a": 1, "b": 2, "c": 3}), Dict))

    def test_get_by_path(self):
        d = Dict(deepcopy(self.data))

        self.assertEqual(d["c"]._value_, self.data["c"])
        self.assertEqual(d["d/e"]._value_, self.data["d"]["e"])
        self.assertEqual(d["d/f"]._value_, self.data["d"]["f"])
        self.assertEqual(d["a/0"]._value_, self.data["a"][0])
        self.assertEqual(d["a/1"]._value_, self.data["a"][1])
        self.assertEqual(d.get("a/1"), self.data["a"][1])
        self.assertEqual(len(d["a"]), 6)

        # self.assertListEqual(list(d["a"][2:6]),       [1.0, 2, 3, 4])

    def test_assign(self):
        cache = {}

        d = Dict(cache)

        d["a"] = "hello world {name}!"

        self.assertEqual(cache["a"], "hello world {name}!")
        d["e"] = {}
        d["e"]["f"] = 5
        d["e"]["g"] = 6

        self.assertEqual(cache["e"]["f"].__value__, 5)
        self.assertEqual(cache["e"]["g"].__value__, 6)

    def test_update(self):
        d = Dict(deepcopy(test_data))

        d._update_({"d": {"g": 5}})

        self.assertEqual(d["d"]["e"]._value_, "{name} is {age}")
        self.assertEqual(d["d"]["f"]._value_, "{address}")
        self.assertEqual(d["d"]["g"]._value_, 5)

    def test_insert(self):
        d0 = Dict(deepcopy(test_data))

        d0._insert_({"a": "hello world {name}!"})
        d0._update_({"d": {"g": 5}})

        self.assertEqual(d0["d"]["e"]._value_, "{name} is {age}")
        self.assertEqual(d0["d"]["f"]._value_, "{address}")
        self.assertEqual(d0["d"]["g"]._value_, 5)

        d1 = List([])

        d1._insert_({"a": [1], "b": 2})

        self.assertEqual(d1[0]["a"][0]._value_, 1)
        self.assertEqual(d1[0]["b"]._value_, 2)

        d1["0/a"].insert(2)

        self.assertEqual(d1[0]["a"]._value_, [1, 2])

    def test_get_by_index(self):
        data = [1, 2, 3, 4, 5]

        d0 = List[int](data)
        # logger.debug(type(d0[0]))
        self.assertIsInstance(d0[0], int)
        self.assertEqual(d0[0], data[0])
        # self.assertListEqual(list(d0[:]), data)

    def test_node_del(self):
        cache = {"a": ["hello world {name}!", "hello world2 {name}!", 1, 2, 3, 4]}

        d = Dict(cache)

        del d["a"]

        self.assertTrue(cache["a"], _undefined_)

    def test_node_insert(self):
        d = Dict[List]({"this_is_a_cache": True})

        d["a"] = "hello world {name}!"
        self.assertEqual(d["a"]._value_, "hello world {name}!")

        d["c"]._insert_(1.23455)
        d["c"]._insert_({"a": "hello world", "b": 3.141567})

        self.assertEqual(d["c"][0]._value_, 1.23455)
        self.assertEqual(d.get("c/0"), 1.23455)
        self.assertEqual(d["c"][1]["b"]._value_, 3.141567)
        self.assertEqual(d.get("c/1/b"), 3.141567)

    def test_type_hint(self):
        d1 = List[HTree]([])
        d1._insert_({"a": 1, "b": 2})

        self.assertIsInstance(d1[0], HTree)

        self.assertEqual(len(d1), 1)
        self.assertEqual(d1[0]["a"]._value_, 1)
        self.assertEqual(d1[0]["b"]._value_, 2)

        data = [1, 2, 3, 4, 5]

        class Foo:
            def __init__(self, v, **kwargs) -> None:
                self.v = v

        d1 = List[Foo](deepcopy(data))

        self.assertIsInstance(d1[2], Foo)
        self.assertEqual(d1[2].v, data[2])


# class TestQuery(unittest.TestCase):
#     # fmt:off
#     data = [
#         {"name": "zhangsan",    "age": 18,  "address": "beijing"},
#         {"name": "lisi",        "age": 19,  "address": "shanghai"},
#         {"name": "wangwu",      "age": 20,  "address": "guangzhou"},
#         {"name": "zhaoliu",     "age": 21,  "address": "shenzhen"},
#     ]
#     # fmt:on

#     def test_iter(self):
#         d0 = AoS(deepcopy(self.data), identifier="name")

#         self.assertListEqual([v.__value__ for v in d0], self.data)

#     def test_slice(self):
#         d0 = AoS(deepcopy(self.data), default_value={"genders": "male"})

#         res = d0[1:4].__value__

#         self.assertListEqual(res, self.data[1:4])

#     def test_query(self):
#         d0 = AoS(deepcopy(self.data), identifier="name")
#         res = d0.get("zhangsan")
#         self.assertDictEqual(res, self.data[0])


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from spdm.core.HTree import AoS, Dict, List
from spdm.core.sp_property import SpTree, sp_property
from spdm.core.TimeSeries import TimeSeriesAoS
from spdm.utils.logger import logger


class Foo(SpTree):
    a: float = sp_property(default_value=4)
    b: float = sp_property()
    c: float = sp_property()


class Goo(SpTree):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    value: float = sp_property()

    foos: List[Foo] = sp_property(default_value={"a": 1, "b": 2, "c": 3})


class Doo(SpTree):

    foo: Foo = sp_property(default_value={"a": 1})

    goo: Goo = sp_property(default_value={"value": 3.14})

    foo_list: List[Foo] = sp_property()

    balaaa = sp_property[Foo](default={"bala": 1})


eq_data = {
    "time": [0.0],
    "vacuum_toroidal_field": {"r0": 6.2, "b0": [-5.3]},
    "code": {"name":  "eq_analyze", },
    "$default_value": {
        "time_slice": {
            "profiles_2d": {"grid": {"dim1": 129, "dim2": 257}},
            "boundary": {"psi_norm": 0.99},
            "coordinate_system": {"grid": {"dim1": 256, "dim2": 128}}
        }}
}


class Mesh(SpTree):
    dim1: int = sp_property()
    dim2: int = sp_property()


class EquilibriumProfiles2d(SpTree):

    grid: Mesh = sp_property()


class EqTimeSlice(SpTree):

    profiles_2d: AoS[EquilibriumProfiles2d] = sp_property()


class Eq(SpTree):

    time: np.ndarray = sp_property()
    time_slice: TimeSeriesAoS[EqTimeSlice] = sp_property()


class TestSpProperty(unittest.TestCase):
    def test_get(self):
        cache = {"foo": {"a": 1234}, }
        d = Doo(cache)

        self.assertFalse(isinstance(cache["foo"], Foo))
        self.assertTrue(isinstance(d.foo, Foo))
        self.assertTrue(isinstance(d._cache["foo"], Foo))

        self.assertTrue(isinstance(d.balaaa, Foo))
        self.assertTrue(isinstance(d._cache["balaaa"], Foo))

        self.assertEqual(d.foo.a, d._cache["foo"].a)

    def test_default_value(self):

        d = Doo()
        self.assertEqual(d.goo.value, 3.14)
        self.assertEqual(d._cache["goo"].value, 3.14)

    def test_get_list(self):

        cache = {"foo_list": [{"a": 1234}, {"b": 1234}, {"c": 1234}, ]}

        d = Doo(cache)

        self.assertFalse(isinstance(cache["foo_list"], Foo))
        self.assertTrue(isinstance(d.foo_list, List))
        # self.assertTrue(isinstance(cache["foo_list"], List))
        self.assertTrue(isinstance(d.foo_list[0], Foo))

        self.assertEqual(d.foo_list[0]["a"], 1234)

    def test_set(self):
        cache = {"foo": {"a": 1234}}
        d = Doo(cache)
        self.assertEqual(cache["foo"]["a"], 1234)
        d.foo.a = 45678.0
        self.assertEqual(d._cache["foo"].a, 45678)

    def test_delete(self):
        cache = {"foo": {"a": 1234}}
        d = Doo(cache)
        self.assertEqual(cache["foo"]["a"], 1234)
        del d.foo
        self.assertTrue("foo" not in cache)

    def test_list_default_child_value(self):
        cache = [{"a": 6}, {"a": 7}, {"a": 8}, {}]
        d = AoS[Foo](cache, default_value={"a": 1, "b": 2, "c": 3})
        self.assertEqual(d[0].a, 6)
        self.assertEqual(d[0].b, 2)
        self.assertEqual(d[0].c, 3)
        self.assertEqual(d[-1].a, 4)
        self.assertEqual(d[-1].b, 2)
        self.assertEqual(d[-1].c, 3)

    def test_sp_data(self):

        eq = Eq(eq_data)

        self.assertTrue(isinstance(eq.time_slice, TimeSeriesAoS))

        self.assertEqual(eq.time_slice[0].profiles_2d[0].grid.dim1,
                         eq_data["$default_value"]["time_slice"]["profiles_2d"]["grid"]["dim1"])


if __name__ == '__main__':
    unittest.main()



import pathlib
from spdm.core.File import File
from spdm.utils.logger import logger

WORKSPACE = "/home/salmon/workspace"  # "/ssd01/salmon_work/workspace/"

DATA_PATH = pathlib.Path(f"{WORKSPACE}/fytok_data/data")

if __name__ == '__main__':

    with File(DATA_PATH/"g063982.04800", mode="r", format="GEQdsk") as fid:
        doc = fid.read()
        eq = doc.dump()

    with File("../output/test.gfile", mode="w", format="GEQdsk") as oid:
        oid.write(eq)

    with File("../output/test.json", mode="w") as oid:
        oid.write(eq)

    with File("../output/test.json", mode="r") as oid:
        logger.debug(oid.read().dump())

    with File("../output/test.yaml", mode="w") as oid:
        oid.write(eq)

    with File("../output/test.yaml", mode="r") as oid:
        logger.debug(oid.read().dump())

    with File("../output/test.h5", mode="w", format="HDF5") as oid:
        oid.write(eq)

    with File("../output/test.h5", mode="r", format="HDF5") as oid:
        logger.debug(oid.read().dump())

    with File("../output/test.nc", mode="w", format="NetCDF") as oid:
        oid.write(eq)

    with File("../output/test.nc", mode="r", format="NetCDF") as oid:
        logger.debug(oid.read().dump())

    # with open("../output/test.json", mode="w") as fp:
    #     d = doc.dump(enable_ndarray=False)
    #     json.dump(d, fp)



"""
Feature:
 * This module defines functions for Uniform Resource Identifier (URI) string,
    following the syntax specifications in RFC 3986.
 * This module support extended 'Path' syntax, supports bracket '[]' in the path.
 TODO (salmon.20190919): support  quoting
"""
import ast
import collections
import collections.abc
import pathlib
import re
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Union
from urllib.parse import parse_qs, urlparse

from .logger import logger

_rfc3986 = re.compile(
    r"^((?P<protocol>[^:/?#]+):)?(//(?P<authority>[^/?#]*))?(?P<path>[^?#]*)(\?(?P<query>[^#]*))?(#(?P<fragment>.*))?")

_rfc3986_ext0 = re.compile(
    r"^((?P<schema>[^:/?#\+\[\]]+)?(\+(?P<protocol>[^:/?#\+\[\]]+))?:)?(//(?P<authority>[^/?#]*))?(?P<path>[^?#]*)(\?(?P<query>[^#]*))?(#(?P<fragment>.*))?")

_rfc3986_ext = re.compile(
    r"^((?P<protocol>[^:/?#\+\[\]]+)?(\+(?P<format>[^:/?#\+\[\]]+))?(\[(?P<schema>[^:/?#\+\[\]]*)\])?:)?(//(?P<authority>[^/?#]*))?(?P<path>[^?#]*)(\?(?P<query>[^#]*))?(#(?P<fragment>.*))?")


@dataclass
class URITuple:
    protocol: str = ""
    authority: str = ""
    path: str = ""
    query: dict = None
    fragment: str = ""


def uri_split_as_dict(uri) -> dict:
    if uri is None:
        uri = ""
    elif isinstance(uri, URITuple):
        return uri.__dict__

    # res = _rfc3986.match(uri).groupdict()

    uri_ = urlparse(uri)

    query = "{" + ','.join([(f'"{k}":"{v[0]}"' if not v[0].isnumeric() else f'"{k}":{v[0]}')
                           for k, v in parse_qs(uri_.query).items()])+"}"
    ast.literal_eval(query)

    if uri_.netloc in ['.', '..']:
        path = uri_.netloc+uri_.path
        netloc = ''
    else:
        path = uri_.path
        netloc = uri_.netloc

    res = dict(
        protocol=uri_.scheme,
        authority=netloc,
        path=path,
        query=ast.literal_eval(query),
        fragment=uri_.fragment
    )
    return res


def uri_split(uri: str | URITuple | Path | None) -> URITuple:
    if isinstance(uri, URITuple):
        return deepcopy(uri)
    elif isinstance(uri, str):
        return URITuple(**uri_split_as_dict(uri))
    elif isinstance(uri, (collections.abc.Sequence, Path)):
        return URITuple(path=uri, query={})
    elif uri is None:
        return URITuple(query={})
    else:
        raise TypeError(uri)


def uri_merge(schema, authority=None, path=None,  query=None, fragment=None):

    return "".join([
        schema+"://" if schema is not None else "",
        (authority or "").strip('/'),
        path or "",
        "?"+str(query) if query is not None else "",
        "#"+str(fragment) if fragment is not None else ""
    ])


def uri_join(base, uri):
    o0 = uri_split(base) if not isinstance(
        base, collections.abc.Mapping) else base
    o1 = uri_split(uri) if not isinstance(uri, collections.abc.Mapping) else uri
    if o1.schema is not None and o1.schema != o0.schema:
        return uri
    elif o1.authority is not None and o1.authority != o0.authority:
        schema = o0.schema
        authority = o1.authority
        path = o1.path
    else:
        schema = o0.schema
        authority = o0.authority

        if o1.path is None or o1.path == '':
            path = o0.path
        elif o1.path is not None and len(o1.path) > 0 and o1.path[0] == '/':
            path = o1.path
        else:
            path = o0.path[:o0.path.rfind('/')]+"/"+o1.path

    return uri_merge(schema, authority, path, o1.query, o1.fragment)


def uridefrag(uri):
    o = uri_split(uri)
    return uri_merge(o.schema, o.authority, o.path, None, None), o.fragment


_r_path_item = re.compile(
    r"([a-zA-Z_\$][^./\\\[\]]*)|\[([+-]?\d*)(?::([+-]?\d*)(?::([+-]?\d*))?)?\]")


class _Empty:
    pass


def _ion(v):
    """if v is not return int(v) else return None"""
    return int(v) if v is not None and v != '' else None


def parse_url_iter(path, with_position=False):
    """ obj : object like dict or list
        path:
            i.e.  a.b.d[23][3:3:4].adf[3]
        try_attr: if true then try to get attribute  when getitem failed

    """

    for m in _r_path_item.finditer(path):
        attr, start, stop, step = m.groups()
        if attr is not None:
            idx = attr
        elif stop is None and step is None:
            idx = _ion(start)
        else:
            idx = slice(_ion(start), _ion(stop), _ion(step))

        if with_position:
            yield idx, m.end()
        else:
            yield idx


def normalize_path_to_list(path, split=True):
    if isinstance(path, str):
        if split:
            path = parse_url_iter(path)
        else:
            path = [path]
    elif not isinstance(path, collections.abc.Sequence):
        path = [path]
    return path


def getitem_by_path(obj, path, *, try_attribute=False, split=True):
    path = normalize_path_to_list(path, split)

    if path is None:
        return obj

    for idx in path:
        if isinstance(obj, tuple):
            if isinstance(idx, str):
                obj = getattr(obj, idx)
            else:
                obj = obj[idx]
        elif isinstance(obj, collections.abc.Mapping):
            obj = obj[idx]
        elif isinstance(obj, collections.abc.Sequence) and \
                (isinstance(idx, int) or isinstance(idx, slice)):
            obj = obj[idx]
        elif try_attribute and isinstance(idx, str) and hasattr(obj, idx):
            obj = getattr(obj, idx)
        else:
            raise IndexError(idx)
    return obj


def setitem_by_path(obj, path: str, data, *, try_attribute=True, split=True):
    path = normalize_path_to_list(path, split)

    if path is None:
        return obj

    prev_idx = None

    for idx in path:
        if prev_idx is None:
            pass
        elif isinstance(obj, tuple):
            if isinstance(prev_idx, str):
                obj = getattr(obj, prev_idx)
            else:
                obj = obj[idx]
        elif isinstance(obj, collections.abc.Mapping) and not isinstance(idx, slice):
            obj = obj.setdefault(prev_idx, {})
        elif isinstance(obj, collections.abc.Sequence):
            obj = obj[prev_idx]
        elif try_attribute and isinstance(prev_idx, str):
            obj = getattr(obj, prev_idx)
        else:
            raise IndexError(f"Insert item error {idx} !")
        prev_idx = idx

    if isinstance(obj, tuple):
        raise AttributeError("Can't set attribute to tuple")
    elif isinstance(obj, collections.abc.MutableMapping) or isinstance(obj, collections.abc.MutableSequence):
        obj[prev_idx] = data
    elif try_attribute and isinstance(prev_idx, str) and hasattr(obj, prev_idx):
        setattr(obj, prev_idx, data)
    else:
        raise IndexError(f"Set item error! {obj} {path}")


def getvalue_r(obj, path):
    return getitem_by_path(obj, path, try_attribute=True)


def setvalue_r(obj, path, data):
    return setitem_by_path(obj, path, data, try_attribute=True)


def pathslit(path: Union[List, str], delimiter='/') -> List:
    if isinstance(path, str):
        path = path.split(delimiter)
    return path




from functools import cached_property
from operator import is_

from spdm.numlib import interpolate, np

from ..core.Function import Function
from spdm.utils.logger import logger
from .Line import Line


@Line.register("bspline_curve")
class BSplineCurve(Line):
    def __init__(self, u, p, *args, is_closed=None, cycle=None, **kwargs) -> None:
        # if len(args) != 2:
        #     raise ValueError(f"Illegal input! len(args)={len(args)}")
        super().__init__(*args, is_closed=is_closed is not None, cycle=cycle)

        self._u = u if u is not None else np.linspace(0, 1.0, len(p[0]))
        self._spl = interpolate.make_interp_spline(self._u, np.c_[tuple(p)], **kwargs)
        if self.is_closed:
            self._spl.extrapolate = 'periodic'

    def inside(self, *x):
        return NotImplemented

    @property
    def points(self):
        return self.map(self._u)

    def point(self, u,  *args, **kwargs):
        return self._spl(u,  *args, **kwargs).T

    def map(self, u, *args, **kwargs):
        r"""
            ..math:: \Phi:\mathbb{R}\rightarrow N
        """
        return self._spl(u, *args, **kwargs).T

    def __call__(self, *args, **kwargs):
        return self.map(*args, **kwargs)

    @cached_property
    def _derivative(self):
        return self._spl.derivative()

    def derivative(self,  *args, **kwargs):
        if len(args) == 0:
            return self._derivative(self._u, **kwargs).T
        else:
            return self._derivative(*args, **kwargs).T

    def pullback(self, func, *args, form=0, **kwargs):
        if len(args) > 0:
            return func(*self.map(*args, **kwargs))

        if form == 0:
            fun = Function(self._u, func(*self.map(self._u)), is_period=self.is_closed)
        elif form == 1:
            fun = Function(self._u, func(*self.map(self._u))*self.dl(self._u), is_period=self.is_closed)
        else:
            raise ValueError()

        return fun



import collections.abc
import typing
from io import BytesIO
from spdm.core.Expression import Expression, Variable
from spdm.core.Functor import Functor
from spdm.core.Function import Function
import matplotlib.pyplot as plt
import numpy as np


from spdm.utils.logger import logger
from spdm.utils.typing import array_type
from spdm.view.View import View

_EXPR_OP_NAME = {
    "negative": "-",
    "add": "+",
    "subtract": "-",
    "multiply": "*",
    "matmul": "*",
    "true_divide": "/",
    "power": "^",
    "equal": "==",
    "not_equal": "!",
    "less": "<",
    "less_equal": "<=",
    "greater": ">",
    "greater_equal": ">=",
    "add": "+",
    "subtract": "-",
    "multiply": "*",
    "matmul": "*",
    "divide": "/",
    "power": "^",
    # "abs": "",
    "positive": "+",
    # "invert": "",
    "bitwise_and": "&",
    "bitwise_or": "|",

    # "bitwise_xor": "",
    # "right_shift": "",
    # "left_shift": "",
    # "right_shift": "",
    # "left_shift": "",
    "mod": "%",
    # "floor_divide": "",
    # "floor_divide": "",
    # "trunc": "",
    # "round": "",
    # "floor": "",
    # "ceil": "",
}
# fmt:off
_EXPR_OP_LATEX = {
    "negative":         ("-",       0),
    "positive":         ("+",       0),
    "add":              ("+",       1),
    "subtract":         ("-",       1),
    "multiply":         ("\\cdot", 2),
    "matmul":           ("\\cdot", 2),
    "divide":           ("/",       2),    
    "true_divide":      ("/",       2),
    "power":            ("^",       3),
    "equal":            ("==",      0),
    "not_equal":        ("!",       0),
    "less":             ("<",       0),
    "less_equal":       ("<=",      0),
    "greater":          (">",       0),
    "greater_equal":    (">=",      0),


    # "abs":            ("",1),

    # "invert":         ("",1),
    "bitwise_and":      ("&",       0),
    "bitwise_or":       ("|",       0),

    # "bitwise_xor": "",
    # "right_shift": "",
    # "left_shift": "",
    # "right_shift": "",
    # "left_shift": "",
    "mod": "%",
    # "floor_divide": "",
    # "floor_divide": "",
    # "trunc": "",
    # "round": "",
    # "floor": "",
    # "ceil": "",
}
# fmt:on


@View.register(["Latex", "latex"])
class LatexView(View):
    backend = "latex"

    def __init__(self, *args,   **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def draw(self, expr: typing.Any, *args, **kwargs) -> str:
        if isinstance(expr, Variable):
            return f"${expr.__label__}$"
        else:
            return f"$${self._render(expr, *args, **kwargs)[0]}$$"

    def _render(self, expr: typing.Any, *args, parent: int = None, **kwargs) -> typing.Tuple[str, int]:

        level = 10

        if isinstance(expr, np.ndarray):
            res = f"<{expr.shape}>"

        elif np.isscalar(expr):
            res = f"{expr}"

        elif isinstance(expr, Variable):
            res = expr.__label__
            level = 10

        elif isinstance(expr, Function):
            res, level = self._render_function(expr, *args, parent=parent, **kwargs)

        elif isinstance(expr, Expression) and isinstance(expr._op, Functor) and isinstance(expr._op._func, np.ufunc):
            res, level = self._render_ufunc(expr, *args, parent=parent, **kwargs)

        elif isinstance(expr, Expression) and isinstance(expr._op, Functor):
            res, level = self._render_expr(expr, *args, parent=parent, **kwargs)

        else:
            res = str(expr)

        if parent is not None and parent > level:
            res = f"\\left({res}\\right)"

        return res, level

    def _render_function(self, func: Function, *args, **kwargs) -> typing.Tuple[str, int]:
        return func.__label__, 0

    def _render_ufunc(self, expr: Expression, *args, **kwargs) -> typing.Tuple[str, int]:
        op: np.ufunc = expr._op._func

        if expr._op._method != "__call__" and expr._op._method is not None:
            raise NotImplementedError(f"op={expr._op._func} method={expr._op._method}")

        children = expr._children
        n_children = len(children)

        tag, level = _EXPR_OP_LATEX.get(op.__name__, (None, 10))

        if n_children != 2 or tag is None:
            res = f"{op.__name__}({', '.join([self._render(child)[0] for child in  children])})"

        elif op.__name__ in ("true_divide", "divide"):
            res = f"\\frac{{{self._render(children[0],parent=level)[0]}}}{{{self._render(children[1])[0]}}}"

        else:
            res = f"{self._render(expr._children[0],parent=level)[0]} {tag} {self._render(expr._children[1],parent=level)[0]} "

        return res, level

    def _render_expr(self, expr: Expression, *args, parent=None, **kwargs) -> typing.Tuple[str, int]:
        return f"{expr._op.__label__}\\left({', '.join([self._render(child, *args, parent=None,**kwargs)[0] for child in  expr._children])}\\right)", 0



from __future__ import annotations
import warnings
import typing
from copy import copy, deepcopy
import functools
import collections.abc
import numpy as np

from ..utils.typing import ArrayType
from ..utils.tags import _not_found_
from ..utils.typing import ArrayType, NumericType, array_type, as_array, is_scalar, is_array, numeric_type
from ..utils.logger import logger
from ..utils.misc import group_dict_by_prefix
from ..numlib.interpolate import interpolate

from .Functor import Functor
from .Entry import Entry
from .HTree import HTreeNode, HTree, HTreeNode, List
from .Domain import DomainBase
from .Path import update_tree, Path
from .Functor import Functor, DerivativeOp


_T = typing.TypeVar("_T", float, bool, array_type, HTreeNode)


class Expression(HTreeNode):
    """Expression

    表达式是由多个操作数和运算符按照约定的规则构成的一个序列。
    其中运算符表示对操作数进行何种操作，而操作数可以是变量、常量、数组或者表达式。
    表达式可以理解为树状结构，每个节点都是一个操作数或运算符，每个节点都可以有多个子节点。
    表达式的值可以通过对树状结构进行遍历计算得到。
    没有子节点的节点称为叶子节点，叶子节点可以是常量、数组，也可以是变量和函数。

    变量是一种特殊的函数，它的值由上下文决定。

    例如：
        >>> import spdm
        >>> x = spdm.core.Expression(op=np.sin)
        >>> y = spdm.core.Expression(op=np.cos)
        >>> z = x + y
        >>> z
        <Expression   op="add" />
        >>> z(0.0)
        3.0


    """

    Domain = DomainBase

    @staticmethod
    def guess_dims(holder, prefix="coordinate", **kwargs):
        if holder is None or holder is _not_found_:
            return _not_found_
        elif isinstance(holder, List):
            return Expression.guess_dims(holder._parent, prefix=prefix, **kwargs)

        coords = []

        metadata = getattr(holder, "_metadata", {})

        dims_s, *_ = group_dict_by_prefix(metadata, prefix, sep=None)

        if dims_s is not None and len(dims_s) > 0:
            dims_s = {int(k): v for k, v in dims_s.items() if k.isdigit()}
            dims_s = dict(sorted(dims_s.items(), key=lambda x: x[0]))

            for c in dims_s.values():
                if not isinstance(c, str):
                    d = as_array(c)
                elif c == "1...N":
                    d = None
                # elif isinstance(holder, HTree):
                #     d = holder.get(c, _not_found_)
                else:
                    d = Path(c).get(holder, _not_found_)

                if d is _not_found_ or d is None:
                    # logger.warning(f"Can not get coordinates {c} from {holder}")
                    coords = []
                    break
                coords.append(as_array(d))

                # elif c.startswith("../"):
                #     d = as_array(holder._parent.get(c[3:], _not_found_))
                # elif c.startswith(".../"):
                #     d = as_array(holder._parent.get(c, _not_found_))
                # elif hasattr(holder.__class__, "get"):
                #     d = as_array(holder.get(c, _not_found_))
                # else:
                #     d = _not_found_
                # elif c.startswith("*/"):
                #     raise NotImplementedError(f"TODO:{self.__class__}.dims:*/")
                # else:
                #     d = as_array(holder.get(c, _not_found_))

        if len(coords) == 0:
            return Expression.guess_dims(getattr(holder, "_parent", None), prefix=prefix, **kwargs)
        else:
            return tuple(coords)

    @staticmethod
    def guess_domain(obj, **kwargs):
        holder = obj
        domain = None
        while holder is not _not_found_ and holder is not None:
            domain_pth = getattr(holder, "_metadata", {}).get("domain", None)
            if domain_pth is not None:
                domain = Path(domain_pth).get(holder, _not_found_)
                if domain is not _not_found_:
                    break
            holder = getattr(holder, "_parent", None)

        if isinstance(domain, array_type) and domain.ndim == 1:
            domain = [domain]

        if domain is None or domain is _not_found_:
            domain = Expression.guess_dims(obj, **kwargs)

        return domain

    def __init__(self, op, *args, domain=_not_found_, **kwargs) -> None:
        """
        Parameters


        op : typing.Callable  | ExprOp
            运算符，可以是函数，也可以是类的成员函数
        args : typing.Any
            操作数
        kwargs : typing.Any
            命名参数， 用于传递给运算符的参数

        """

        if self.__class__ is not Expression:
            super().__init__(**kwargs)
            self._op = op
            self._children = args
            self._domain = domain

        elif is_scalar(op):  # 常量/标量表达式
            match op:
                case 0:
                    TP = ConstantZero
                case 1:
                    TP = ConstantOne
                case _:
                    TP = Scalar

            self.__class__ = TP
            TP.__init__(self, op, *args, domain=domain, **kwargs)

        elif isinstance(op, array_type):  # 构建插值函数
            from .Function import Function

            TP = Function

            self.__class__ = TP
            TP.__init__(self, op, *args, domain=domain, **kwargs)

        else:  # 一般表达式
            super().__init__(**kwargs)
            self._op = op
            self._children = args
            self._domain = domain

    def __copy__(self) -> Expression:
        """复制一个新的 Expression 对象"""
        other: Expression = super().__copy__()
        other._op = copy(self._op)
        other._domain = self._domain
        other._children = copy(self._children)
        return other

    def __serialize__(self, dumper=None):
        logger.debug(f"TODO: __serialize__ {self}")
        return None

    @property
    def domain(self) -> DomainBase:
        """返回表达式的定义域"""

        if self._domain is None or self._domain is _not_found_:
            self._domain = Expression.guess_domain(self)

        if self._domain is None or self._domain is _not_found_:
            for child in self._children:
                domain = getattr(child, "domain", _not_found_)
                if domain is not _not_found_:
                    break
                elif isinstance(domain, DomainBase) and not domain.is_full:
                    break
            else:
                domain = _not_found_

            self._domain = domain

        if self._domain is _not_found_:
            raise RuntimeError(f"Can not get domain!")

        elif not isinstance(self._domain, DomainBase):
            self._domain = self.__class__.Domain(self._domain)

        return self._domain

    @property
    def has_children(self) -> bool:
        """判断是否有子节点"""
        return len(self._children) > 0

    @property
    def empty(self) -> bool:
        return not self.has_children and self._op is None

    def __repr__(self) -> str:
        return self._render_latex_()

    def __str__(self) -> str:
        return self._render_latex_()

    def _render_latex_(self) -> str:
        vargs = []
        for expr in self._children:
            if isinstance(expr, (bool, int, float, complex)):
                res = str(expr)
            elif expr is None:
                res = "n.a"
            elif isinstance(expr, np.ndarray):
                if len(expr.shape) == 0:
                    res = f"{expr.item()}"
                else:
                    res = f"{expr.dtype}[{expr.shape}]"
            elif isinstance(expr, Variable):
                res = f"{expr.__label__}"
            elif isinstance(expr, Expression):
                res = expr._render_latex_()
            elif isinstance(expr, np.ufunc):
                res = expr.__name__
            else:
                res = expr.__class__.__name__
            vargs.append(res)

        if isinstance(self._op, np.ufunc):
            op_tag = EXPR_OP_TAG.get(self._op.__name__, self._op.__name__)

            if self._op.nin == 1:
                res = rf"{op_tag}{{{vargs[0]}}}"

            elif self._op.nin == 2:
                if op_tag == "/":
                    res = f"\\frac{{{vargs[0]}}}{{{vargs[1]}}}"
                else:
                    res = rf"{{{vargs[0]}}} {op_tag} {{{vargs[1]}}}"
            else:
                raise RuntimeError(f"Tri-op is not defined!")

        elif (op_tag := self._metadata.get("label", None) or self._metadata.get("name", None)) is not None:
            if len(vargs) == 0:
                res = op_tag
            else:
                res: str = rf"{op_tag}\left({','.join(vargs)}\right)"

        else:
            if isinstance(self._op, Expression):
                op_tag = self._op.__label__
            else:
                op_tag = self._op.__class__.__name__

            res: str = rf"{op_tag}\left({','.join(vargs)}\right)"

        return res

    def _repr_latex_(self) -> str:
        """for jupyter notebook display"""
        return f"$${self._render_latex_()}$$"

    @property
    def dtype(self):
        return float

    def __array_ufunc__(self, ufunc, method, *args, **kwargs) -> Expression:
        """
        重载 numpy 的 ufunc 运算符, 用于在表达式中使用 numpy 的 ufunc 函数构建新的表达式。
        例如：
            >>> import numpy as np
            >>> import spdm
            >>> x = spdm.core.Expression(np.sin)
            >>> y = spdm.core.Expression(np.cos)
            >>> z = x + y
            >>> z
            <Expression   op="add" />
            >>> z(0.0)
            1.0
        """
        if method != "__call__" or len(kwargs) > 0:
            return Expression(Functor(ufunc, method=method, **kwargs), *args)
        else:
            return Expression(ufunc, *args)

    def __array__(self, dtype=None) -> array_type:
        """在定义域上计算表达式，返回数组。"""

        if self._cache is None or self._cache is _not_found_:
            if self.domain is None or self.domain is _not_found_:
                raise RuntimeError(f"Domain is not defined")

            x = self.domain.points

            if len(x) == 0:
                raise RuntimeError(f"Empty domain, __array__ fail!")

            self._cache = self.__call__(*x)

        return self._cache

    def __ppoly__(self):
        """表达式的多项式近似

        Raises:
            RuntimeError: 当 domain 为空时报错

        Returns:
            多项式: 多项式
        """
        if getattr(self, "_ppoly", _not_found_) is _not_found_:
            self._ppoly = self.domain.interpolate(self.__array__())

        return self._ppoly

    def _eval_children(self, *args, **kwargs):
        if len(self._children) == 0:
            return args
        else:
            new_children = []
            for child in self._children:
                try:
                    if callable(child) and len(args) + len(kwargs) > 0:
                        value = child(*args, **kwargs)
                    else:
                        value = np.asarray(child)

                except Exception as error:
                    raise RuntimeError(f"Failure to calculate  child {child} !") from error
                else:
                    new_children.append(value)
            return new_children

    def __eval__(self, *args, **kwargs):
        if not callable(self._op):
            raise RuntimeError(f"Unknown functor {self._op} {type( self._op)}")

        new_children = self._eval_children(*args, **kwargs)

        # 执行当前节点算符
        res = self._op(*new_children)

        # with warnings.catch_warnings():
        #     warnings.filterwarnings("error", category=RuntimeWarning)
        #     try:
        #         # 执行当前节点算符
        #         res = self._op(*new_children)
        #     except RuntimeWarning:
        #         logger.exception(f"{self._render_latex_()} {self._op} {new_children}")
        #         # raise RuntimeError((res, args))
        #         # res=np.nan_to_num(res,nan=1.0e-33)
        #         res = np.nan

        return res

    def __call__(self, *args, **kwargs) -> Expression | array_type:
        """
        重载函数调用运算符，用于计算表达式的值

        TODO:
        - support JIT compilation
        - support broadcasting?
        - support multiple meshes?

        Parameters

        xargs : NumericType
            自变量/坐标，可以是标量，也可以是数组
        kwargs : typing.Any
            命名参数，用于传递给运算符的参数
        """
        if len(args) + len(kwargs) == 0:  # 空调用，返回自身
            return self

        elif any([isinstance(a, Expression) for a in args]):  #  创建复合函数
            return Expression(self, *args, **kwargs)

        else:  # 执行计算
            return self.__eval__(*args, **kwargs)

    def derivative(self, order: int, *args, **kwargs) -> Derivative:
        return Derivative(self, *args, order=order, **kwargs)

    def antiderivative(self, order: int, *args, **kwargs) -> Antiderivative:
        return Antiderivative(self, *args, order=order, **kwargs)

    def partial_derivative(self, order: typing.Tuple[int, ...], *args, **kwargs) -> PartialDerivative:
        return PartialDerivative(self, *args, order=order, **kwargs)

    def pd(self, *order, **kwargs) -> PartialDerivative:
        return self.partial_derivative(order, **kwargs)

    def integral(self, *args, **kwargs) -> float:
        raise NotImplementedError(f"")

    @property
    def d(self) -> Expression:
        """1st derivative 一阶导数"""
        return self.derivative(1)

    @property
    def d2(self) -> Expression:
        """2nd derivative 二阶导数"""
        return self.derivative(2)

    @property
    def I(self) -> Expression:
        """antiderivative 原函数"""
        return self.antiderivative(1)

    @property
    def dln(self) -> Expression:
        """logarithmic derivative 对数求导"""
        return self.derivative(1) / self

    def find_roots(self, *args, **kwargs) -> typing.Generator[float, None, None]:
        raise NotImplementedError(f"TODO: find_roots")

    def fetch(self, *args, _parent=None, **kwargs):
        if len(args) + len(kwargs) == 0:
            if self._cache is not None:
                return self._cache
            else:
                return self.__array__()
        else:
            res = self.__call__(*args, **kwargs)

            if res is self:
                res = self.__copy__()

            if isinstance(res, HTreeNode):
                res._parent = _parent

        return res

    # fmt: off
    def __neg__      (self                             ) : return Expression(np.negative     ,  self     )
    def __add__      (self, o: NumericType | Expression) : return Expression(np.add          ,  self, o  ) if not ((is_scalar(o) and o == 0 ) or isinstance(o, ConstantZero) or o is _not_found_ and o is None) else self
    def __sub__      (self, o: NumericType | Expression) : return Expression(np.subtract     ,  self, o  ) if not ((is_scalar(o) and o == 0 ) or isinstance(o, ConstantZero) or o is _not_found_ and o is None) else self
    def __mul__      (self, o: NumericType | Expression) : return Expression(np.multiply     ,  self, o  ) if not (is_scalar(o) and (o ==0 or o==1)) else (ConstantZero() if o==0 else self)
    def __matmul__   (self, o: NumericType | Expression) : return Expression(np.matmul       ,  self, o  ) if not (is_scalar(o) and (o ==0 or o==1)) else (ConstantZero() if o==0 else self)
    def __truediv__  (self, o: NumericType | Expression) : return Expression(np.true_divide  ,  self, o  ) if not (is_scalar(o) and (o ==0 or o==1)) else (Scalar(np.nan) if o==0 else self)
    def __pow__      (self, o: NumericType | Expression) : return Expression(np.power        ,  self, o  ) if not (is_scalar(o) and (o ==0 or o==1)) else (ConstantOne() if o==0 else self)
    def __eq__       (self, o: NumericType | Expression) : return Expression(np.equal        ,  self, o  )
    def __ne__       (self, o: NumericType | Expression) : return Expression(np.not_equal    ,  self, o  )
    def __lt__       (self, o: NumericType | Expression) : return Expression(np.less         ,  self, o  )
    def __le__       (self, o: NumericType | Expression) : return Expression(np.less_equal   ,  self, o  )
    def __gt__       (self, o: NumericType | Expression) : return Expression(np.greater      ,  self, o  )
    def __ge__       (self, o: NumericType | Expression) : return Expression(np.greater_equal,  self, o  )
    def __radd__     (self, o: NumericType | Expression) : return Expression(np.add          ,  o, self  ) if not ((is_scalar(o) and o == 0 ) or isinstance(o, ConstantZero) or o is _not_found_ and o is None) else self
    def __rsub__     (self, o: NumericType | Expression) : return Expression(np.subtract     ,  o, self  ) if not ((is_scalar(o) and o == 0 ) or isinstance(o, ConstantZero) or o is _not_found_ and o is None) else self.__neg__()
    def __rmul__     (self, o: NumericType | Expression) : return Expression(np.multiply     ,  o, self  ) if not (is_scalar(o) and (o ==0 or o==1)) else (ConstantZero() if o==0 else self)
    def __rmatmul__  (self, o: NumericType | Expression) : return Expression(np.matmul       ,  o, self  ) if not (is_scalar(o) and (o ==0 or o==1)) else (ConstantZero() if o==0 else self)
    def __rtruediv__ (self, o: NumericType | Expression) : return Expression(np.divide       ,  o, self  )
    def __rpow__     (self, o: NumericType | Expression) : return Expression(np.power        ,  o, self  ) if not (is_scalar(o) and o ==1)  else ConstantOne()
    def __abs__      (self                             ) : return Expression(np.abs          ,  self     )
    def __pos__      (self                             ) : return Expression(np.positive     ,  self     )
    def __invert__   (self                             ) : return Expression(np.invert       ,  self     )
    def __and__      (self, o: NumericType | Expression) : return Expression(np.bitwise_and  ,  self, o  ) if not isinstance(o,bool) else ( self if o ==True else False)
    def __or__       (self, o: NumericType | Expression) : return Expression(np.bitwise_or   ,  self, o  ) if not isinstance(o,bool) else ( True if o ==True else self)
    def __xor__      (self, o: NumericType | Expression) : return Expression(np.bitwise_xor  ,  self, o  )
    def __rand__     (self, o: NumericType | Expression) : return Expression(np.bitwise_and  ,  o, self  ) if not isinstance(o,bool) else ( self if o ==True else False)
    def __ror__      (self, o: NumericType | Expression) : return Expression(np.bitwise_or   ,  o, self  ) if not isinstance(o,bool) else ( True if o ==True else self)
    def __rxor__     (self, o: NumericType | Expression) : return Expression(np.bitwise_xor  ,  o, self  )
    def __rshift__   (self, o: NumericType | Expression) : return Expression(np.right_shift  ,  self, o  )
    def __lshift__   (self, o: NumericType | Expression) : return Expression(np.left_shift   ,  self, o  )
    def __rrshift__  (self, o: NumericType | Expression) : return Expression(np.right_shift  ,  o, self  )
    def __rlshift__  (self, o: NumericType | Expression) : return Expression(np.left_shift   ,  o, self  )
    def __mod__      (self, o: NumericType | Expression) : return Expression(np.mod          ,  self, o  )
    def __rmod__     (self, o: NumericType | Expression) : return Expression(np.mod          ,  o, self  )
    def __floordiv__ (self, o: NumericType | Expression) : return Expression(np.floor_divide ,  self, o  )
    def __rfloordiv__(self, o: NumericType | Expression) : return Expression(np.floor_divide ,  o, self  )
    def __trunc__    (self                             ) : return Expression(np.trunc        ,  self     )
    def __round__    (self, n=None                     ) : return Expression(np.round        ,  self, n  )
    def __floor__    (self                             ) : return Expression(np.floor        ,  self     )
    def __ceil__     (self                             ) : return Expression(np.ceil         ,  self     )
    # fmt: on


EXPR_OP_TAG = {
    "negative": "-",
    "add": "+",
    "subtract": "-",
    "multiply": r"\cdot",
    "matmul": r"\cdot",
    "true_divide": "/",
    "power": "^",
    "equal": "==",
    "not_equal": "!",
    "less": "<",
    "less_equal": "<=",
    "greater": ">",
    "greater_equal": ">=",
    "add": "+",
    "subtract": "-",
    "multiply": r"\cdot",
    "matmul": r"\cdot",
    "divide": "/",
    "power": "^",
    # "abs": "",
    "positive": "+",
    # "invert": "",
    "bitwise_and": "&",
    "bitwise_or": "|",
    # "bitwise_xor": "",
    # "right_shift": "",
    # "left_shift": "",
    # "right_shift": "",
    # "left_shift": "",
    "mod": "%",
    # "floor_divide": "",
    # "floor_divide": "",
    # "trunc": "",
    # "round": "",
    # "floor": "",
    # "ceil": "",
    "sqrt": r"\sqrt",
}


class Variable(Expression):
    """Variable

    变量是一种特殊的函数，它的值由上下文决定。
    例如：
    >>> import spdm
    >>> x = spdm.core.Variable(0,"x")
    >>> y = spdm.core.Variable(1,"y")
    >>> z = x + y
    >>> z
    <Expression   op="add" />
    >>> z(0.0, 1.0)
    1.0

    """

    def __init__(self, idx: int | str, name: str = None, domain=None, **kwargs) -> None:
        if name is None:
            name = idx if isinstance(idx, str) else f"_{idx}"
        super().__init__(None, name=name, domain=domain, **kwargs)
        self._idx = idx

    def __copy__(self) -> Scalar:
        res = super().__copy__()
        res._idx = self._idx
        return res

    @property
    def _type_hint(self) -> typing.Type:
        """获取函数的类型"""
        orig_class = getattr(self, "__orig_class__", None)
        if orig_class is not None:
            return typing.get_args(orig_class)[0]
        else:
            return float

    @property
    def index(self):
        return self._idx

    def __call__(self, *args, **kwargs):
        if all([isinstance(a, Variable) for a in args]):
            res = self
        elif len(args) == 1 and isinstance(args[0], HTree):
            pth = Path(self.__name__)
            res = pth.get(args[0], _not_found_)
            if res is _not_found_:
                res = pth.get(kwargs, _not_found_)
            if res is _not_found_:
                raise RuntimeError(f"Can not get variable {self.__name__}")
        elif self._idx < len(args):
            res = args[self._idx]
        elif self.__name__.isidentifier():
            res = kwargs.get(self.__name__)
        else:
            raise RuntimeError(
                f"Variable {self.__label__} require {self._idx+1} args, or {self.__name__} in kwargs but only {args} provided!"
            )

        return res

    def __repr__(self) -> str:
        return self.__label__


class Scalar(Expression):
    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 0 or not isinstance(args[0], (float, int, bool)):
            raise ValueError(f"args should be float|int|bool")
        super().__init__(None, *args, **kwargs)

    @property
    def __label__(self):
        return self._children[0]

    def __array__(self) -> ArrayType:
        return np.array(self._children[0])

    def __str__(self):
        return str(self._children[0])

    def __repr__(self) -> str:
        return str(self._children[0])

    def __equal__(self, other) -> bool:
        return self._children[0] == other

    def __call__(self, *args, **kwargs):
        return self._children[0]

    def derivative(self, *args, **kwargs):
        return ConstantZero(_parent=self._parent, **kwargs)


class ConstantZero(Scalar):
    def __init__(self, *args, **kwargs):
        super().__init__(0, **kwargs)

    # fmt: off
    def __neg__      (self                             ) : return self
    def __add__      (self, o: NumericType | Expression) : return o
    def __sub__      (self, o: NumericType | Expression) : return Expression(np.negative     ,  o  ) 
    def __mul__      (self, o: NumericType | Expression) : return self
    def __matmul__   (self, o: NumericType | Expression) : return self
    def __truediv__  (self, o: NumericType | Expression) : return self
    def __pow__      (self, o: NumericType | Expression) : return self
    def __radd__     (self, o: NumericType | Expression) : return o
    def __rsub__     (self, o: NumericType | Expression) : return o
    def __rmul__     (self, o: NumericType | Expression) : return self
    def __rmatmul__  (self, o: NumericType | Expression) : return self
    def __rtruediv__ (self, o: NumericType | Expression) : return Scalar(np.nan)
    def __rpow__     (self, o: NumericType | Expression) : return one
    def __abs__      (self                             ) : return self
    # fmt: on


class ConstantOne(Scalar):
    def __init__(self, *args, **kwargs):
        super().__init__(1, **kwargs)

    # fmt: off
    def __neg__      (self                             ) : return Scalar(-1)
    def __mul__      (self, o: NumericType | Expression) : return o
    def __matmul__   (self, o: NumericType | Expression) : return o
    def __pow__      (self, o: NumericType | Expression) : return self
    def __rmul__     (self, o: NumericType | Expression) : return o
    def __rmatmul__  (self, o: NumericType | Expression) : return o
    def __rtruediv__ (self, o: NumericType | Expression) : return o
    def __rpow__     (self, o: NumericType | Expression) : return o
    def __abs__      (self                             ) : return self
    # fmt: on


zero = ConstantZero()

one = ConstantOne()


class Derivative(Expression):
    """算符: 用于表示一个运算符，可以是函数，也可以是类的成员函数
    受 np.ufunc 启发而来。
    可以通过 ExprOp(op, method=method) 的方式构建一个 ExprOp 对象。

    """

    def __init__(self, *args, order=1, **kwargs):
        if len(args) == 0:
            raise RuntimeError(f"Need at least one position argument!")
        elif len(args) == 1:
            expr = args[0]
            args = []
        elif all([isinstance(a, array_type) for a in args]):
            y, *x = args
            expr = Expression(*x, y)
            args = []
        else:
            expr = None
        super().__init__(None, *args, **kwargs)

        self._order = order
        self._expr = expr

    @property
    def order(self) -> int | None:
        return self._order

    @property
    def domain(self) -> DomainBase:
        domain = super().domain
        if (domain is None or domain is _not_found_) and isinstance(self._expr, Expression):
            domain = self._expr.domain
        return domain

    def _render_latex_(self) -> str:
        expr: Expression = self._expr

        if expr is None or expr is _not_found_:
            return self.__label__

        match self._order:
            case 0:
                text = expr._render_latex_()
            case 1:
                text = f"d{expr._render_latex_()}"
            case -1:
                text = rf"\int \left({expr._render_latex_()} \right)"
            case -2:
                text = rf"\iint \left({expr._render_latex_()} \right)"
            case _:
                if self._order > 1:
                    text = rf"d_{{\left[{self._order}\right]}}{expr._render_latex_()}"
                elif self._order < 0:
                    text = rf"\intop^{{{-self._order}}}\left({expr._render_latex_()}\right)"
                else:
                    text = expr._render_latex_()
        return text

    def __eval__(self, *args: _T, **kwargs) -> _T:
        if len(self._children) > 0:
            y, *x = self._eval_children(*args)
            ppoly = interpolate(*x, y, **kwargs)

            if self._order > 0:
                res = ppoly.derivative(self._order)(*x)
            elif self._order < 0:
                res = ppoly.antiderivative(-self._order)(*x)
            else:
                res = ppoly(*x)

            return res

        elif self._op is None or self._op is _not_found_:
            try:
                ppoly = self._expr.__ppoly__()
            except Exception:
                y = self._expr(*args)
                x = args
                ppoly = interpolate(*x, y)

            if self._order > 0:
                self._op = ppoly.derivative(self._order)
            elif self._order < 0:
                self._op = ppoly.antiderivative(-self._order)
            else:
                self._op = ppoly

        return self._op(*args, **kwargs)


class Antiderivative(Derivative):
    def __init__(self, *args, order=1, **kwargs):
        super().__init__(*args, order=-order, **kwargs)


def derivative(*args, order=1, **kwargs):
    func = Derivative(*args, order=order, **kwargs)
    if all([isinstance(d, (array_type, float, int)) for d in args[1:]]):
        _, *x = args
        return func(*x)
    else:
        return func


def antiderivative(*args, order=1, **kwargs):
    return Antiderivative(*args, order=order, **kwargs)


class PartialDerivative(Derivative):
    def __repr__(self) -> str:
        return f"d_{{{self._order}}} ({Expression._repr_s(self._expr)})"


from ..numlib.smooth import smooth as _smooth

from scipy.signal import savgol_filter


class SmoothOp(Expression):
    def __init__(self, op, *args, **kwargs) -> None:
        super().__init__(op or savgol_filter, *args, options=kwargs)

    def __eval__(self, y: array_type, *args, **kwargs) -> typing.Any:
        if len(args) + len(kwargs) > 0:
            logger.warning(f"Ignore {args} {kwargs}")

        return self._op(y, **self._metadata.get("options", {}))

    def __call__(self, *args, **kwargs):
        if len(args) + len(kwargs) > 0:
            return super().__call__(*args, **kwargs)
        else:
            return self.__eval__(*self._children)


def smooth(expr, *args, op=None, **kwargs):
    if isinstance(expr, array_type):
        return SmoothOp(op, expr, *args, **kwargs)()
    else:
        return SmoothOp(op, expr, *args, **kwargs)




import typing
import unittest
from copy import deepcopy

import numpy as np

from spdm.utils.logger import logger

from spdm.geometry.GeoObject import GeoObject, GeoObject1D, GeoObject2D, GeoObject3D


class TestGeoObject(unittest.TestCase):

    def test_define_new_class(self):

        class GObj(GeoObject1D):

            @property
            def rank(self) -> int: return 10

        obj = GObj(np.zeros((10, 10)))

        self.assertEqual(obj.rank, 10)


if __name__ == '__main__':
    unittest.main()



import unittest
from copy import deepcopy

from spdm.utils.logger import logger
from spdm.utils.tags import _not_found_
from spdm.core.Entry import Entry, as_value


class TestEntry(unittest.TestCase):
    data = {
        "a": [
            "hello world {name}!",
            "hello world2 {name}!",
            1, 2, 3, 4
        ],
        "c": "I'm {age}!",
        "d": {
            "e": "{name} is {age}",
            "f": "{address}"
        }
    }

    def test_get(self):

        d = Entry(self.data)

        self.assertEqual(d.get("c"),           self.data["c"])
        self.assertEqual(d.get("d/e"),      self.data["d"]["e"])
        self.assertEqual(d.get("d/f"),      self.data["d"]["f"])
        self.assertEqual(d.get("a/0"),        self.data["a"][0])
        self.assertEqual(d.get("a/1"),        self.data["a"][1])

    def test_exists(self):
        d = Entry(self.data)

        self.assertTrue(d.exists)
        self.assertTrue(d.child("a").exists)
        self.assertTrue(d.child("d/e").exists)
        self.assertFalse(d.child("b/h").exists)
        self.assertFalse(d.child("f/g").exists)

    def test_count(self):
        d = Entry(self.data)

        self.assertEqual(d.count,          1)
        self.assertEqual(d.child("a").count,   6)
        self.assertEqual(d.child("d").count,   1)

    def test_insert(self):
        cache = deepcopy(self.data)
        d = Entry(cache)

        d.child("c").insert(1.23455)

        d.child("c").insert([{"a": "hello world", "b": 3.141567}])

        self.assertEqual(cache["c"][1],                      1.23455)
        self.assertEqual(cache["c"][2]["a"],           "hello world")
        self.assertEqual(cache["c"][2]["b"],                3.141567)

    def test_update(self):
        cache = deepcopy(self.data)

        d = Entry(cache)

        d.update({"d": {"g": 5}})

        self.assertEqual(cache["d"]["e"], "{name} is {age}")
        self.assertEqual(cache["d"]["f"], "{address}")
        self.assertEqual(cache["d"]["g"], 5)

    def test_erase(self):
        cache = {
            "a": [
                "hello world {name}!",
                "hello world2 {name}!",
                1, 2, 3, 4
            ],
            "b": "hello world!"
        }

        d = Entry(cache)
        del d["b"]
        self.assertTrue("b" not in cache)

    def test_get_many(self):
        cache = deepcopy(self.data)

        d = Entry(cache)

        self.assertEqual(d.child('a/2').__value__, self.data['a'][2])

        res = d.get({"a/2", "c",  "d/e", "e"})

        self.assertDictEqual(res, {"a/2": cache['a'][2],
                                   "c": cache['c'],
                                   "d/e":   cache['d']['e'],
                                   "e": _not_found_})

    def test_iter(self):
        data = [1, 2, 3, 4, 5]

        d0 = Entry(data)

        self.assertListEqual([v for v in d0.for_each()], data)

    def test_find_next(self):
        data = [1, 2, 3, 4, 5]

        d0 = Entry(data)

        res = []

        next_id = []

        while True:
            value, next_id = d0.find_next(*next_id)
            if next_id is None or len(next_id) == 0:
                break
            res.append(value)

        self.assertListEqual(res, data)


if __name__ == '__main__':
    unittest.main()
